#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 17:01

from pyxllib.prog.pupil import check_install_package

# 这个需要C++14编译器 https://download.microsoft.com/download/5/f/7/5f7acaeb-8363-451f-9425-68a90f98b238/visualcppbuildtools_full.exe
# 在需要的时候安装，防止只是想用pyxllib很简单的功能，但是在pip install阶段处理过于麻烦
# MatchSimString计算编辑距离需要
check_install_package('Levenshtein', 'python-Levenshtein')

from collections import defaultdict
from more_itertools import chunked
import warnings

import Levenshtein
import numpy as np
import pandas as pd

from pyxllib.prog.pupil import run_once
from pyxllib.prog.specialist import dataframe_str
from pyxllib.text.pupil import briefstr

# 忽略特定的警告
warnings.filterwarnings("ignore", category=FutureWarning,
                        module="sklearn.cluster._agglomerative",
                        lineno=1005)


@run_once('str')
def get_levenshtein_similar(x, y):
    """ 缓存各字符串之间的编辑距离 """
    return Levenshtein.ratio(x, y)


class MatchSimString:
    """ 匹配近似字符串

    mss = MatchSimString()

    # 1 添加候选对象
    mss.append_candidate('福州+厦门2018初数暑假讲义-请录入-快乐学习\初一福厦培优-测试用')
    mss.append_candidate('2018_快乐数学_六年级_秋季_第01讲_圆柱与圆锥_教案（教师版）')
    mss.append_candidate('删除所有标签中间多余的空白')

    # 2 需要匹配的对象1
    s = '奕本初一福周厦门培油'

    idx, sim = mss.match(s)
    print('匹配目标：', mss[idx])  # 匹配目标： 福州+厦门2018初数暑假讲义-请录入-快乐学习\初一福厦培优-测试用
    print('相似度：', sim)         # 相似度： 0.22

    # 3 需要匹配的对象2
    s = '圆柱与【圆锥】_教案空白版'

    idx, sim = mss.match(s)
    print('匹配目标：', mss[idx])  # 2018_快乐数学_六年级_秋季_第01讲_圆柱与圆锥_教案（教师版）
    print('相似度：', sim)         # 相似度： 0.375

    如果append_candidate有传递2个扩展信息参数，可以索引获取：
    mss.ext_value[idx]
    """

    def __init__(self, method=briefstr):
        self.preproc = method
        self.origin_str = []  # 原始字符串内容
        self.key_str = []  # 对原始字符串进行处理后的字符
        self.ext_value = []  # 扩展存储一些信息

    def __getitem__(self, item):
        return self.origin_str[item]

    def __delitem__(self, item):
        del self.origin_str[item]
        del self.key_str[item]
        del self.ext_value[item]

    def __len__(self):
        return len(self.key_str)

    def get_similarity(self, x, y):
        """ 计算两对数据之间的相似度 """
        pass

    def append_candidate(self, k, v=None):
        self.origin_str.append(k)
        if callable(self.preproc):
            k = self.preproc(k)
        self.key_str.append(k)
        self.ext_value.append(v)

    def match(self, s):
        """ 跟候选字符串进行匹配，返回最佳匹配结果
        """
        idx, sim = -1, 0
        for i in range(len(self)):
            k, v = self.key_str[i], self.ext_value[i]
            sim_ = Levenshtein.ratio(k, s)
            if sim_ > sim:
                sim = sim_
                idx = i
            i += 1
        return idx, sim

    def match_many(self, s, count=1):
        """跟候选字符串进行匹配，返回多个最佳匹配结果
        :param str s: 待匹配的字符串
        :param int count: 需要返回的匹配数量
        :return: 匹配结果列表，列表中的元素为(idx, sim)对
        """
        scores = [(i, Levenshtein.ratio(self.key_str[i], s)) for i in range(len(self))]
        # 根据相似度排序并返回前count个结果
        return sorted(scores, key=lambda x: x[1], reverse=True)[:count]

    def match_test(self, s, count=-1, showstr=lambda x: x[:50]):
        """输入一个字符串s，和候选项做近似匹配

        :param s: 需要进行匹配的字符串s
        :param count: 只输出部分匹配结果
            -1：输出所有匹配结果
            0 < count < 1：例如0.4，则只输出匹配度最高的40%结果
            整数：输出匹配度最高的count个结果
        :param showstr: 字符串显示效果
        """
        # 1 计算编辑距离，存储结果到res
        res = []
        n = len(self)
        for i in range(n):
            k, v = self.key_str[i], self.ext_value[i]
            sim = Levenshtein.ratio(k, s)
            res.append([i, v, sim, showstr(k)])  # 输出的时候从0开始编号
            i += 1

        # 2 排序、节选结果
        res = sorted(res, key=lambda x: -x[2])
        if 0 < count < 1:
            n = max(1, int(n * count))
        elif isinstance(count, int) and count > 0:
            n = min(count, n)
        res = res[:n]

        # 3 输出
        df = pd.DataFrame.from_records(res, columns=('序号', '标签', '编辑距离', '内容'))
        s = dataframe_str(df)
        s = s.replace('\u2022', '')  # texstudio无法显示会报错的字符
        print(s)

    def agglomerative_clustering(self, threshold=0.5):
        """ 对内部字符串进行层次聚类

        :param threshold: 可以理解成距离的阈值，距离小于这个阈值的字符串会被聚为一类
            值越小，分出的类别越多越细
        """
        check_install_package('sklearn', 'scikit-learn')
        from sklearn.cluster import AgglomerativeClustering

        # 1 给每个样本标类别
        distance_matrix = np.zeros((len(self), len(self)))
        for i in range(len(self)):
            for j in range(i + 1, len(self)):
                # 我们需要距离，所以用1减去相似度
                distance = 1 - Levenshtein.ratio(self.key_str[i], self.key_str[j])
                distance_matrix[i, j] = distance_matrix[j, i] = distance

        # 进行层次聚类
        clustering = AgglomerativeClustering(n_clusters=None, affinity='precomputed',
                                             distance_threshold=threshold,
                                             linkage='complete')
        labels = clustering.fit_predict(distance_matrix)

        return labels

    def display_clusters(self, threshold=0.5):
        """ 根据agglomerative_clustering的结果，显示各个聚类的内容 """

        labels = self.agglomerative_clustering(threshold=threshold)
        cluster_dict = defaultdict(list)

        # 组织数据到字典中
        for idx, label in enumerate(labels):
            cluster_dict[label].append(self.origin_str[idx])

        # 按标签排序并显示
        result = {}
        for label, items in sorted(cluster_dict.items(), key=lambda x: -len(x[1])):
            result[label] = items

        return result


class HierarchicalMatchSimString(MatchSimString):
    """ 在面对数据量很大的候选数据情况下，建议使用这个层次聚类后的匹配方法 """

    def __init__(self, method=briefstr):
        super().__init__(method)
        self.groups = dict()

    def get_center_sample(self, indices=None):
        """ 输入一组下标，计算中心样本，未输入参数值的时候，则在全量样本里找 """
        if indices is None:
            indices = range(len(self))

        # 用于存储之前计算的结果
        cached_results = {}

        def get_similarity(i, j):
            """ 获取两个索引的相似度，利用缓存来避免重复计算 """
            if (i, j) in cached_results:
                return cached_results[(i, j)]
            sim_val = Levenshtein.ratio(self.key_str[i], self.key_str[j])
            cached_results[(i, j)] = cached_results[(j, i)] = sim_val
            return sim_val

        center_idx = max(indices, key=lambda x: sum(get_similarity(x, y) for y in indices))
        return center_idx

    def merge_group(self, indices, threshold=0.5, strategy='center'):
        """ 对输入的indexs清单，按照threshold的阈值进行合并
        返回的是一个字典，key是代表性样本，value是同组内的数据编号

        :param strategy: 代表样本的挑选策略
            center，中心样本
            first，第一个样本
        """
        check_install_package('sklearn', 'scikit-learn')
        from sklearn.cluster import AgglomerativeClustering

        # 1 给每个样本标类别
        n = len(indices)
        distance_matrix = np.zeros((n, n))
        for i in range(n):
            for j in range(i + 1, n):
                # 我们需要距离，所以用1减去相似度
                distance = 1 - Levenshtein.ratio(self.key_str[indices[i]], self.key_str[indices[j]])
                distance_matrix[i, j] = distance_matrix[j, i] = distance

        # 进行层次聚类
        clustering = AgglomerativeClustering(n_clusters=None, affinity='precomputed',
                                             distance_threshold=threshold,
                                             linkage='complete')
        labels = clustering.fit_predict(distance_matrix)

        # 2 分组字典
        cluster_dict = defaultdict(list)
        # 组织数据到字典中
        for i, label in enumerate(labels):
            cluster_dict[label].append(indices[i])

        # 3 改成代表样本映射到一组里，并且按照样本数从多到少排序
        result = {}
        for label, items in sorted(cluster_dict.items(), key=lambda x: -len(x[1])):
            if strategy == 'first':
                representative = items[0]
            elif strategy == 'center':
                # 使用局部索引计算平均距离
                local_indices = [i for i, idx in enumerate(indices) if idx in items]
                sub_matrix = distance_matrix[np.ix_(local_indices, local_indices)]
                avg_distances = sub_matrix.mean(axis=1)
                representative_idx = np.argmin(avg_distances)
                representative = items[representative_idx]
            else:
                raise ValueError(f'Invalid strategy: {strategy}')
            result[representative] = items

        return result

    def init_groups(self, threshold=0.5, batch_size=1000):
        """
        :param threshold: 按照阈值进行分组，在这个距离内的都会归到一组
        :param batch_size: 因为数据可能太大，不可能一次性全量两两比较，这里可以分batch处理
            这样虽然结果不太精确，但能大大减小运算量
        """
        # 1 最开始每个样本都是一个组
        groups = {i: [i] for i in range(len(self))}
        new_groups = {}

        # 2 不断合并，直到没有组数变化
        while len(groups) > 1:
            for indices in chunked(groups.keys(), batch_size):
                # 对于这里返回的字典，原groups里的values也要对应拼接的
                indices2 = self.merge_group(indices, threshold=threshold)
                for idx, idxs in indices2.items():
                    # 获取原始分组中的索引
                    original_idxs = [groups[original_idx] for original_idx in idxs]
                    # 展平列表并分配到新分组中
                    new_groups[idx] = [item for sublist in original_idxs for item in sublist]

            # 如果分组没有发生变化，退出循环
            if len(new_groups) == len(groups):
                break

            groups = new_groups
            new_groups = {}

        # 3 按数量从多到少排序
        new_groups = {}
        for label, items in sorted(groups.items(), key=lambda x: -len(x[1])):
            new_groups[label] = items  # 暂用第一个出现的作为代表

        self.groups = new_groups
        return self.groups



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com

"""
字符分类
"""

from collections import Counter
import math
import re

from pyxllib.prog.pupil import run_once, safe_div

# todo 加个 公式类型？
charclass_dict = {
    "标点": """－：、，；。！？“”（）【】""",
    "常用汉字": "一丁七万丈三上下不与丐丑专且世丘丙业丛东丝丢两严丧个中丰串临丸丹为主丽举乃久么义之乌乍乎乏乐乒乓乔乖乘乙九乞也习乡书买乱乳干了予争事二于亏云互五井亚些亡交亥亦产亩享京亭亮亲人亿什仁仅仆仇今介仍从仑仓仔他仗付仙代令以仪们仰仲件价任份仿企伊伍伏伐休众优伙会伞伟传伤伦伪伯估伴伶伸伺似佃但位低住佑体何余佛作你佣佩佳使侄侈例侍供依侠侣侥侦侧侨侮侯侵便促俄俊俏俐俗俘保信俩俭修俯俱俺倍倒倔倘候倚借倡倦债值倾假偎偏做停健偶偷偿傀傅傍储催傲傻像僚僧僵僻儒儡儿允元兄充兆先光克免兑兔党兜兢入全八公六兰共关兴兵其具典养兼兽冀内冈册再冒冕冗写军农冠冤冬冯冰冲决况冶冷冻净凄准凉凌减凑凛凝几凡凤凫凭凯凰凳凶凸凹出击函凿刀刁刃分切刊刑划列刘则刚创初删判刨利别刮到制刷券刹刺刻刽剂剃削前剑剔剖剥剧剩剪副割剿劈力劝办功加务劣动助努劫励劲劳势勃勇勉勋勒勘募勤勺勾勿匀包匆匈匕化北匙匠匣匪匹区医匾匿十千升午半华协卑卒卓单卖南博卜占卡卢卤卦卧卫印危即却卵卷卸卿厂厅历厉压厌厕厘厚原厢厦厨去县叁参又叉及友双反发叔取受变叙叛叠口古句另叨只叫召叭叮可台史右叶号司叹叼叽吁吃各吆合吉吊同名后吏吐向吓吕吗君吝吞吟吠否吧吨吩含听吭吮启吱吴吵吸吹吻吼呀呆呈告呐呕员呛呜呢周味呵呻呼命咆和咏咐咒咕咖咙咧咨咪咬咱咳咸咽哀品哄哆哈响哎哑哗哟哥哨哩哪哭哮哲哺哼唁唆唇唉唐唠唤唧唬售唯唱唾啃啄商啊啡啤啥啦啰啸啼喂善喇喉喊喘喜喝喧喳喷喻嗅嗓嗜嗡嗤嗦嗽嘀嘁嘉嘱嘲嘴嘶嘹嘿器噩噪嚎嚣嚷嚼囊囚四回因团囤园困囱围固国图圃圆圈土圣在地场圾址均坊坎坏坐坑块坚坛坝坞坟坠坡坤坦坪坯坷垂垃垄型垒垛垢垦垫垮埂埃埋城域埠培基堂堆堕堡堤堪堰堵塌塑塔塘塞填境墅墓墙增墨墩壁壕壤士壮声壳壶壹处备复夏夕外多夜够大天太夫夭央夯失头夷夸夹夺奄奇奈奉奋奏契奔奕奖套奠奢奥女奴奶奸她好如妄妆妇妈妒妓妖妙妥妨妹妻姆姊始姐姑姓委姚姜姥姨姻姿威娃娄娇娘娜娩娱娶婆婉婚婴婶婿媒媚媳嫁嫂嫉嫌嫡嫩嬉子孔孕字存孙孝孟季孤学孩孵孽宁它宅宇守安宋完宏宗官宙定宛宜宝实宠审客宣室宦宪宫宰害宴宵家容宽宾宿寂寄密寇富寒寓寝寞察寡寥寨寸对寺寻导寿封射将尉尊小少尔尖尘尚尝尤就尸尺尼尽尾尿局屁层居屈屉届屋屎屏屑展属屠屡履屯山屹屿岁岂岔岖岗岛岩岭岳岸峡峦峭峰峻崇崎崔崖崩崭嵌巍川州巡巢工左巧巨巩巫差己已巴巷巾币市布帅帆师希帐帕帖帘帚帜帝带席帮常帽幅幌幔幕幢平年并幸幻幼幽广庄庆庇床序庐库应底店庙府庞废度座庭庵庶康庸廉廊廓延廷建开异弃弄弊式弓引弛弟张弥弦弧弯弱弹强归当录形彤彩彪彬彭彰影役彻彼往征径待很徊律徐徒得徘徙御循微德徽心必忆忌忍志忘忙忠忧快忱念忽忿怀态怎怒怔怕怖怜思怠急性怨怪怯总恃恋恍恐恒恕恢恤恨恩恬恭息恰恳恶恼悄悉悍悔悟悠患悦您悬悯悲悴悼情惊惋惑惕惜惠惦惧惨惩惫惭惯惰想惶惹愁愈愉意愕愚感愤愧愿慈慌慎慕慢慧慨慰慷憋憎憔憨憾懂懈懊懒懦戈戏成我戒或战戚截戳戴户房所扁扇手才扎扑扒打扔托扛扣执扩扫扬扭扮扯扰扳扶批扼找承技抄把抑抒抓投抖抗折抚抛抠抡抢护报披抬抱抵抹押抽拂拄担拆拇拉拌拍拐拒拓拔拖拗拘拙招拜拟拢拣拥拦拧拨择括拭拯拱拳拴拷拼拾拿持挂指按挎挑挖挚挟挠挡挣挤挥挨挪挫振挺挽捂捅捆捉捌捍捎捏捐捕捞损捡换捣捧据捶捷捺捻掀掂授掉掌掏掐排掖掘掠探接控推掩措掰掷掸掺揉揍描提插揖握揣揩揪揭援揽搀搁搂搅搏搓搔搜搞搪搬搭携摄摆摇摊摔摘摧摩摸摹撇撑撒撕撞撤撩撬播撮撰撵撼擂擅操擎擒擦攀攒攘支收改攻放政故效敌敏救教敛敞敢散敦敬数敲整敷文斋斑斗料斜斟斤斥斧斩断斯新方施旁旅旋族旗无既日旦旧旨早旬旭旱时旷旺昂昆昌明昏易昔昙星映春昧昨昭是昵昼显晃晋晌晒晓晕晚晤晦晨普景晰晴晶智晾暂暇暑暖暗暮暴曙曲更曹曼曾替最月有朋服朗望朝期朦木未末本术朱朴朵机朽杀杂权杆杈杉李杏材村杖杜束杠条来杨杭杯杰松板极构枉析枕林枚果枝枢枣枪枫枯架枷柄柏某柑柒染柔柜柠查柬柱柳柴柿栅标栈栋栏树栓栖栗校株样核根格栽桂桃桅框案桌桐桑档桥桦桨桩桶梁梅梆梗梢梦梧梨梭梯械梳检棉棋棍棒棕棘棚棠森棱棵棺椅植椎椒椭椰椿楔楚楞楣楷楼概榄榆榔榕榛榜榨榴槐槽樊樟模横樱橄橘橙橡橱檀檐檩檬欠次欢欣欧欲欺款歇歉歌止正此步武歧歪歹死歼殃殉殊残殖殴段殷殿毁毅母每毒比毕毙毛毡毫毯氏民氓气氛氢氧氨氮氯水永汁求汇汉汗汛汞江池污汤汪汰汹汽沃沈沉沐沙沛沟没沥沦沧沪沫沮河沸油治沼沽沾沿泄泉泊泌法泛泞泡波泣泥注泪泰泳泵泻泼泽洁洋洒洗洛洞津洪洲活洼洽派流浅浆浇浊测济浑浓浙浦浩浪浮浴海浸涂消涉涌涎涕涛涝涡涣涤润涧涨涩涮涯液涵淀淆淋淌淑淘淡淤淫淮深淳混淹添清渊渐渔渗渠渡渣渤温港渴游渺湃湖湘湾湿溃溅溉源溜溢溪溯溶溺滋滑滓滔滚滞满滤滥滨滩滴漂漆漏漓演漠漩漫漱漾潘潜潦潭潮澄澈澎澜澡澳激濒瀑灌火灭灯灰灵灶灸灼灾灿炉炊炎炒炕炫炬炭炮炸点炼烁烂烈烘烙烛烟烤烦烧烫热烹焊焕焙焚焦焰然煌煎煞煤照煮熄熊熏熔熙熟熬燃燎燕燥爆爪爬爱爵父爷爸爹爽片版牌牍牙牛牡牢牧物牲牵特牺犀犁犬犯状犹狂狈狐狗狞狠狡独狭狮狰狱狸狼猎猖猛猜猩猪猫猬献猴猾猿玄率玉王玖玛玩玫环现玲玷玻珊珍珠班球琅理琉琐琢琳琴琼瑞瑟瑰璃璧瓜瓢瓣瓤瓦瓮瓶瓷甘甚甜生甥用甩甫田由甲申电男甸画畅界畏畔留畜略畦番畴畸疆疏疑疗疙疚疟疤疫疮疯疲疹疼疾病症痊痒痕痘痛痢痪痰痴痹瘟瘤瘦瘩瘪瘫瘸瘾癌癞癣登白百皂的皆皇皮皱皿盅盆盈益盏盐监盒盔盖盗盘盛盟目盯盲直相盹盼盾省眉看真眠眨眯眶眷眼着睁睛睡督睦睬睹瞄瞎瞒瞧瞪瞬瞳瞻矗矛矢知矩矫短矮石矾矿码砂砌砍研砖砚砰破砸砾础硅硕硝硫硬确硼碉碌碍碎碑碗碘碟碧碰碱碳碴碾磁磅磕磨磷磺礁示礼社祈祖祝神祟祠祥票祭祷祸禀禁福离禽禾秀私秃秆秉秋种科秒秕秘租秤秦秧秩秫积称秸移秽稀程稍税稚稠稳稻稼稽稿穆穗穴究穷空穿突窃窄窍窑窒窖窗窘窜窝窟窥窿立竖站竞竟章竣童竭端竹竿笆笋笑笔笙笛笤符笨第笼等筋筏筐筑筒答策筛筝筷筹签简箍箕算管箩箫箭箱篇篓篙篡篮篱篷簇簸簿籍米类籽粉粒粗粘粟粤粥粪粮粱粹精糊糕糖糙糜糟糠糯系紊素索紧紫累絮繁纠红纤约级纪纫纬纯纱纲纳纵纷纸纹纺纽线练组绅细织终绊绍绎经绑绒结绕绘给络绝绞统绢绣继绩绪续绰绳维绵绷绸综绽绿缀缅缆缎缓缔缕编缘缚缝缠缤缨缩缭缰缴缸缺罐网罕罗罚罢罩罪置署羊美羔羞羡群羹羽翁翅翎翔翘翠翩翰翻翼耀老考者而耍耐耕耗耘耙耳耸耻耽耿聂聊聋职联聘聚聪肃肄肆肉肋肌肖肘肚肛肝肠股肢肤肥肩肪肮肯育肴肺肾肿胀胁胃胆背胎胖胚胜胞胡胧胯胰胳胶胸能脂脆脉脊脏脐脑脓脖脚脯脱脸脾腊腋腌腐腔腕腥腮腰腹腺腻腾腿膀膊膏膘膛膜膝膨膳臀臂臊臣自臭至致臼舀舅舆舌舍舒舔舞舟航般舰舱舵舶舷船艇艘良艰色艳艺艾节芋芍芒芙芜芝芥芦芬芭芯花芳芹芽苇苍苏苔苗苛苞苟若苦苫英苹茁茂范茄茅茉茎茧茫茬茴茵茶茸荆草荐荒荔荚荞荠荡荣荤荧药荷荸莉莫莱莲获莹莺莽菇菊菌菜菠菩菱菲萄萌萍萎萝萤营萧萨落著葛葡董葫葬葱葵蒂蒋蒙蒜蒲蒸蒿蓄蓉蓖蓝蓬蔑蔓蔗蔚蔫蔬蔼蔽蕉蕊蕴蕾薄薇薛薪薯藏藐藕藤藻蘑蘸虎虏虐虑虚虫虱虹虽虾蚀蚁蚂蚊蚌蚓蚕蚜蚣蚤蚪蚯蛀蛆蛇蛉蛋蛔蛙蛛蛤蛮蛹蛾蜀蜂蜈蜒蜓蜕蜗蜘蜜蜡蜻蝇蝉蝌蝎蝗蝙蝠蝴蝶螃融螟螺蟀蟆蟋蟹蠕蠢血衅行衍衔街衙衡衣补表衩衫衬衰衷袁袄袋袍袒袖袜被袭袱裁裂装裆裕裙裤裳裸裹褂褐褒褥褪襟西要覆见观规觅视览觉角解触言誉誊誓警譬计订认讥讨让训议讯记讲讳讶许讹论讼讽设访诀证评诅识诈诉诊词译试诗诚话诞诡询该详诫诬语误诱诲说诵请诸诺读诽课谁调谅谆谈谊谋谍谎谐谒谓谚谜谢谣谤谦谨谬谭谱谴谷豁豆豌象豪豫豹豺貌贝贞负贡财责贤败账货质贩贪贫贬购贮贯贰贱贴贵贷贸费贺贼贾贿赁赂赃资赊赋赌赎赏赐赔赖赘赚赛赞赠赡赢赤赦赫走赴赵赶起趁超越趋趟趣足趴趾跃跋跌跑跛距跟跨跪路跳践跷跺踊踏踢踩踪踱蹂蹄蹈蹋蹦蹬蹭蹲躁躏身躬躯躲躺车轧轨轩转轮软轰轴轻载轿较辅辆辈辉辐辑输辕辖辙辛辜辞辟辣辨辩辫辰辱边辽达迁迂迄迅过迈迎运近返还这进远违连迟迫述迷迹追退送适逃逆选逊透逐递途逗通逛逝逞速造逢逮逸逻逼逾遂遇遍遏道遗遣遥遭遮遵避邀邑邓邢那邦邪邮邻郁郊郎郑部郭都鄙酌配酒酗酝酣酥酪酬酱酵酷酸酿醇醉醋醒采释里重野量金鉴针钉钓钙钝钞钟钠钢钥钦钧钩钮钱钳钻钾铁铃铅铆铐铛铜铝铡铣铭铲银铸铺链销锁锄锅锈锉锋锌锐错锚锡锣锤锥锦锨锭键锯锰锹锻镀镇镊镐镜镣镰镶长门闪闭问闯闰闲间闷闸闹闺闻闽阀阁阅阎阐阔队阱防阳阴阵阶阻阿附际陆陈陋陌降限陕陡院除陨险陪陵陶陷隅隆随隐隔隘隙障隧隶难雀雁雄雅集雇雌雏雕雨雪雳零雷雹雾需震霉霍霎霜霞露霸霹青靖静非靠靡面革靴靶鞋鞍鞠鞭韧韩韭音韵页顶顷项顺须顽顾顿颁颂预颅领颇颈颊频颓颖颗题颜额颠颤风飒飘飞食餐饥饭饮饰饱饲饵饶饺饼饿馁馅馆馋馍馏馒首香马驮驯驰驱驳驴驶驹驻驼驾骂骄骆骇验骏骑骗骚骡骤骨髓高鬓鬼魁魂魄魏魔鱼鲁鲜鲤鲫鲸鳄鳍鳖鳞鸟鸠鸡鸣鸥鸦鸭鸯鸳鸵鸽鸿鹃鹅鹉鹊鹏鹤鹦鹰鹿麦麸麻黄黍黎黑黔默鼎鼓鼠鼻齐齿龄龙龟",
    "繁体汉字": "丟丣並乗乹乾亁亂亙亝亞亱亷亾佇佈佔併來侖侶侷俁係俓俔俠俥俬俻倀倆倈倉個倐們倖倣倫倲倸偉偑側偵偽傌傑傖傘備傚傢傭傯傳傴債傷傾僂僅僉僊働僑僕僞僤僥僨僱價儀儁儂億儈儉儌儎儐儔儕儘償儣優儭儲儷儸儺儻儼兇兌兎兒兗兠內兩冄冊冐冑冪冺凈凍凙凜凢凱凴別刦刧刪刼剄則剉剋剎剏剗剙剛剝剮剳剴創剷剹剾劃劄劇劉劊劌劍劏劑劒劚効勁勅勌勑動務勛勝勞勢勣勦勩勱勳勵勸勻匟匭匯匱匲匳區協卹卻卽厀厙厠厤厭厰厲厴參叄叢吒吚吳吶呂呌呪咊咼員哯哶唄唓唕唘唚唸啎問啑啓啗啞啟啢啣喎喚喪喫喬單喲嗁嗆嗇嗊嗎嗚嗩嗰嗶嗹嘆嘍嘑嘓嘔嘖嘗嘜嘠嘩嘪嘮嘯嘰嘳嘵嘷嘸嘺嘽噁噅噉噓噚噝噞噠噥噦噯噲噴噸噹嚀嚇嚌嚐嚕嚙嚛嚥嚦嚧嚨嚮嚲嚳嚴嚶嚽囀囁囂囃囅囈囉囌囑囒囓囙囪圅圇國圍園圓圖團圞坿垜垻埡埨埬埰執堅堊堖堚堝堯報場塊塋塏塒塗塚塟塢塤塲塵塸塹塿墊墖墜墠墮墰墲墳墶墻墾壇壈壋壎壓壗壘壙壚壜壞壟壠壢壣壩壪壯壺壻壼壽夘夠夢夥夾奐奧奩奪奬奮奼妝妬妳妷姉姍姙姦姪姸娙娛婁婡婣婦婬婭媈媍媧媯媰媼媽媿嫋嫗嫰嫵嫺嫻嫿嬀嬃嬇嬈嬋嬌嬙嬝嬡嬣嬤嬦嬪嬭嬰嬸嬻嬾孃孄孆孇孋孌孎孫學孻孼孾孿宂宮寀寃寑寠寢實寧審寫寬寳寵寶尅將專尋對導尒尙尟尠尷屆屍屓屛屜屢層屨屩屬屭岅岡峝峯峴島峽崍崑崗崙崢崬嵐嵗嵼嵽嵾嶁嶃嶄嶇嶈嶔嶗嶘嶠嶢嶧嶨嶮嶸嶹嶺嶼嶽巊巋巒巔巖巗巘巠巰巵巹帀帋帥師帬帳帶幀幃幇幑幓幗幘幙幚幝幟幣幩幫幬幹幾庫庻庽廁廂廄廈廎廐廕廚廝廞廟廠廡廢廣廧廩廬廳廵廹廻弒弔弳張強彃彄彆彈彌彎彔彙彞彠彥彫彲彿徃後徑從徠徧復徵徹徿怱怳恆恠恡恥悅悞悤悵悶悽惏惡惥惱惲惷惻愛愜愨愴愷愻愽愾慄態慍慘慙慚慟慣慤慪慫慮慳慴慶慺慼慽慾憂憇憊憍憐憑憒憖憚憢憤憫憮憲憶憸憹懀懇應懌懍懎懞懟懣懤懧懨懲懶懷懸懺懼懾戀戇戔戞戧戩戰戱戲戶戹扞抝拋拚挩挱挵挾捄捨捫捱捲掃掄掆掗掙掚掛採掽揀揑揚換揫揮揯揷揹搆搇搉損搖搗搤搥搨搯搵搶搾摃摋摐摑摜摟摯摳摶摺摻撈撊撏撐撓撝撟撡撣撥撦撧撫撲撳撻撾撿擁擄擇擊擋擓擔擕據擟擠擣擥擧擫擬擯擰擱擲擴擷擺擻擼擽擾攄攆攋攏攔攖攙攛攜攝攢攣攤攩攪攬攷敂敍敎敓敗敘敵數敺斂斃斅斆斕斬斷斸於旂旣旤旹旾昇昬時晉晛晝暈暉暐暘暢暫暱曄曆曇曉曊曏曖曠曡曥曨曬書會朞朢朥朧朮朶東杴枱枴柵柺査栁栞栢栰桒桮桱桺桿梔梖梘梜條梟梲棄棊棖棗棟棡棧棲棶椏椗椲椶椷椾楇楊楓楥楨業極榘榦榪榮榲榿構槍槓槕槤槧槨槫槮槳槶槼樁樂樅樑樓標樞樠樢樣樤樧樫樳樸樹樺樿橈橋橜機橢橫橯檁檉檔檜檟檢檣檭檮檯檳檵檸檻櫃櫅櫈櫍櫓櫚櫛櫝櫞櫟櫠櫥櫧櫨櫪櫫櫬櫱櫳櫸櫻欄欅欇權欍欏欐欑欒欓欖欘欝欞欵欽歎歐歛歟歡歲歴歷歸歿殀殘殞殢殤殨殫殭殮殯殰殲殺殻殼毀毆毊毧毬毿氂氈氊氌氣氫氬氭氳氷氾汎汙汚決沒沖況泝洩洶浹浿涇涖涗涼淒淚淛淥淨淩淪淵淶淺渖渙減渢渦測渾湊湋湞湧湯湼溈準溝溡溫溮溳溼滄滅滌滎滙滛滬滯滲滷滸滻滾滿漁漊漍漚漢漣漬漲漵漸漿潁潄潑潔潕潙潚潛潣潤潯潰潷潿澀澁澅澆澇澐澗澠澤澦澩澫澬澮澱澾濁濃濄濆濇濕濘濚濛濜濟濤濧濫濰濱濶濺濼濾濿瀂瀃瀅瀆瀇瀉瀋瀏瀕瀘瀝瀟瀠瀦瀧瀨瀰瀲瀾灃灄灋灍灑灒灕灘灙灝灠灡灣灤灧灩災為烏烖烴無煇煉煑煒煗煙煢煥煩煬煱熂熅熈熉熌熒熓熗熚熡熰熱熲熾燀燁燄燈燉燒燖燙燜營燦燬燭燴燶燻燼燾爃爄爇爍爐爖爗爛爥爧爭爲爺爾牀牆牋牎牐牓牕牘牠牴牽犖犛犞犢犧狀狥狹狽猂猌猙猨猶猻獁獃獄獅獊獎獘獧獨獩獪獫獮獰獱獲獵獷獸獺獻獼玀玁玅珼現琖琱琺琿瑇瑋瑒瑣瑤瑩瑪瑯瑲瑻瑽璉璊璕璗璝璡璢璣璦璫璯環璵璸璼璽璾璿瓄瓅瓊瓏瓔瓕瓚瓛甌" + "賠議結訴關確該總額義訟計與請舊經辦處萬統絕進達負責駁發間項魯閣繪連絡電話說複認細",
    "偏僻汉字": "丂丄丅丆丌丏丒丕丗丞丠両丨丩丫丬丮丯丱丳丵丶丷丼丿乀乁乂乄乆乇乊乑乕乚乛乜乢乣乤乥乧乨乩乪乫乬乭乮乯乲乴乵乶乷乸乺乻乼乽乿亀亃亄亅亇亊亍亐亓亖亗亘亜亟亠亢亣亨亪亯亰亳亴亵亶亸亹亻亼亽仂仃仄仈仉仌仏仐仒仕仚仛仜仝仞仟仠仡仢仦仧仨仩仫仭仮仯仱仳仴仵仸仹仺仼仾伀伂伃伄伅伆伇伈伉伋伌伎伒伓伔伕伖伛伜伝伡伢伣伥伧伨伩伫伬伭伮伱伲伳伵伷伹伻伽伾伿佀佁佂佄佅佉佊佋佌佐佒佖佗佘佚佝佞佟佡佢佤佥佦佧佨佪佫佬佭佮佯佰佱佲佴佶佷佸佹佺佻佼佽佾侀侁侂侃侅侇侉侊侌侎侏侐侑侒侓侔侕侗侘侙侚侜侞侟価侢侤侩侪侫侬侭侰侱侲侳侴侸侹侺侻侼侽侾俀俅俆俇俈俉俋俌俍俎俑俒俕俖俙俚俛俜俞俟俢俣俤俦俧俨俪俫俰俲俳俴俵俶俷俸俹俼俽俾俿倁倂倃倄倅倇倊倌倎倏倓倕倗倛倜倝倞倠倢値倥倧倨倩倪倬倭倮倯倰倱倳倴倵倶倷倹倻倽倿偀偁偂偃偄偅偆偈偊偋偌偍偐偒偓偔偕偖偗偘偙偛偝偞偟偠偡偢偣偤偦偧偨偩偪偫偬偭偮偯偰偱偲偳偸偹偺偻偼偾傁傂傃傄傆傇傈傉傊傋傎傏傐傒傓傔傕傗傛傜傝傞傟傠傡傣傤傥傦傧傩傪傫傮傰傱傶傸傹傺傼傽傿僀僁僃僄僆僇僈僋僌僎僐僒僓僔僖僗僘僙僛僜僝僟僠僡僢僣僦僩僪僫僬僭僮僯僰僲僳僴僶僷僸僺僼僽僾僿儃儅儆儇儊儋儍儏儑儓儖儗儙儚儛儜儝儞儠儢儤儥儦儧儨儩儫儬儮儯儰儱儳儴儵儶儹儽儾兀兂兊兏児兓兕兖兘兙兛兝兞兟兡兣兤兦兪兮兯兲兹兺兾兿冁冂冃円冇冉冋冎冏冓冔冖冘冚冝冞冟冡冢冣冥冦冧冨冩冫冭冮冱冴冸冹冼冽冾冿凁凂凃凅凇凊凋凎凐凒凓凔凕凖凗凘凚凞凟凣凥処凧凨凩凪凬凮凲凵凷凼凾刂刄刅刈刉刋刌刍刎刏刐刓刔刕刖刜刞刟刡刢刣刬刭刯刱刲刳刴刵刾刿剀剁剅剆剈剌剐剒剓剕剘剚剜剞剟剠剡剢剣剤剦剨剫剬剭剰剱剶剸剺剻剼剽劀劁劂劅劆劋劎劐劓劔劕劖劗劘劙劜劢劤劥劦劧劬劭劮劯劰労劵劶劷劸劺劻劼劽劾勀勂勄勆勈勊勍勎勏勐勓勔勖勗勚勜勠勡勥勧勨勪勫勬勭勮勯勰勲勴勶勷勹勼勽匁匂匃匄匇匉匊匋匌匍匎匏匐匑匒匓匔匘匚匛匜匝匞匡匢匤匥匦匧匨匩匫匬匮匰匴匵匶匷匸匼匽卂卄卅卆卉卋卌卍卐単卙卛卝卞卟卣卥卨卩卪卬卭卮卯卲卶卺卼卾厁厃厄厇厈厊厍厎厏厐厑厒厓厔厖厗厛厜厝厞厡厣厥厧厩厪厫厬厮厯厱厳厵厶厷厸厹厺厼厽厾叀叅叆叇収叏叐叒叓叕叚叜叝叞叟叡叧叩叱叴叵叺叻叾叿吀吂吅吇吋吔吖吘吙吜吡吢吣吤吥吪吰吲吷吺吽吾吿呁呃呄呅呇呉呋呍呎呏呑呒呓呔呖呗呙呚呝呞呟呠呡呣呤呥呦呧呩呫呬呭呮呯呰呱呲呴呶呷呸呹呺呾呿咀咁咂咃咄咅咇咈咉咋咍咎咑咓咔咗咘咚咛咜咝咞咟咠咡咢咣咤咥咦咩咫咭咮咯咰咲咴咵咶咷咹咺咻咾咿哂哃哅哇哉哊哋哌哏哐哒哓哔哕哖哘哙哚哛哜哝哞哠哢哣哤哦哧哫哬哰哱哳哴哵哷哸哹哻哽哾哿唀唂唃唅唈唊唋唌唍唎唏唑唒唔唖唗唙唛唜唝唞唟唡唢唣唥唦唨唩唪唫唭唰唲唳唴唵唶唷唹唺唻唼唽唿啀啁啂啅啇啈啉啋啌啍啐啒啔啕啖啘啙啚啛啜啝啠啧啨啩啪啫啬啭啮啯啱啲啳啴啵啶啷啹啺啻啽啾啿喀喁喃喅喆喈喋喌喍喏喐喑喒喓喔喕喖喗喙喛喞喟喠喡喢喣喤喥喦喨喩喭喯喰喱喴喵営喸喹喺喼喽喾喿嗀嗂嗃嗄嗈嗉嗋嗌嗍嗏嗐嗑嗒嗔嗕嗖嗗嗘嗙嗛嗝嗞嗟嗠嗢嗣嗥嗧嗨嗪嗫嗬嗭嗮嗯嗱嗲嗳嗴嗵嗷嗸嗺嗻嗼嗾嗿嘂嘃嘄嘅嘇嘈嘊嘋嘌嘎嘏嘐嘒嘕嘘嘙嘚嘛嘝嘞嘟嘡嘢嘣嘤嘥嘦嘧嘨嘫嘬嘭嘻嘼嘾噀噂噃噄噆噇噈噊噋噌噍噎噏噐噑噒噔噕噖噗噘噙噛噜噟噡噢噣噤噧噫噬噭噮噰噱噳噵噶噷噺噻噼噽噾噿嚁嚂嚃嚄嚅嚆嚈嚉嚊嚋嚍嚏嚑嚒嚓嚔嚖嚗嚘嚚嚜嚝嚞嚟嚠嚡嚢嚤嚩嚪嚫嚬嚭嚯嚰嚱嚵嚸嚹嚺嚻嚾嚿囄囆囇囋囍囎囏囐囔囕囖囗囘囜囝囟囡団囥囦囧囨囩囫囬囮囯囲図囵囶囷囸囹囻囼囿圀圁圂圄圉圊圌圎圏圐圑圔圕圗圙圚圛圜圝圠圡圢圤圥圦圧圩圪圫圬圭圮圯圱圲圳圴圵圶圷圸圹圻圼圽圿坁坂坃坄坅坆坈坉坋坌坍坒坓坔坕坖坘坙坜坢坣坥坧坨坩坫坬坭坮坰坱坲坳坴坵坶坸坹坺坻坼坽坾垀垁垅垆垇垈垉垊垌垍垎垏垐垑垓垔垕垖垗垘垙垚垝垞垟垠垡垣垤垥垧垨垩垪垬垭垯垰垱垲垳垴垵垶垷垸垹垺垼垽垾垿埀埁埄埅埆埇埈埉埊埌埍埏埐埑埒埓埔埕埖埗埘埙埚埛埜埝埞埢埣埤埥埦埧埩埪埫埭埮埯埱埲埳埴埵埶埸埻埼埽埾埿堀堁堃堄堇堈堉堋堌堍堎堏堐堑堒堓堔堗堘堙堛堜堞堟堠堢堣堥堦堧堨堩堫堬堭堮堲堳堶堷堸堹堺堻堼堽堾堿塀塁塂塃塄塅塆塇塈塉塍塎塐塓塕塖塙塛塜塝塠塡塣塥塦塧塨塩塪塬塭塮塯塰塱塳塴塶塷塺塻塼塽塾墀墁墂墄墆墇墈墉墋墌墍墎墏墐墑墒墔墕増墘墚墛墝墟墡墢墣墤墥墦墧墪墫墬墭墯墱墴墵墷墸墹墺墼墽墿壀壂壃壄壅壆壉壊壌壍壏壐壑壒壔壖壛壝壡壥壦壧壨壬壭壱売壴壵壷壸壾壿夀夁夂夃夅夆夈変夊夋夌夎夐夑夒夓夔夗夙夛夝夞夡夣夤夦夨夬夰夲夳夵夶夻夼夽夿奀奁奂奃奅奆奊奌奍奎奒奓奘奙奚奛奜奝奞奟奡奣奤奦奨奫奭奯奰奱奲奵奷奺奻奾奿妀妁妃妅妉妊妋妌妍妎妏妐妑妔妕妗妘妚妛妜妞妟妠妡妢妣妤妦妧妩妪妫妭妮妯妰妱妲妴妵妶妸妺妼妽妾妿姀姁姂姃姄姅姇姈姌姎姏姒姕姖姗姘姛姝姞姟姠姡姢姣姤姧姩姫姬姭姮姯姰姱姲姳姴姵姶姷姹姺姼姽姾娀娂娅娆娈娉娊娋娌娍娎娏娐娑娒娓娔娕娖娗娚娝娞娟娠娡娢娣娤娥娦娧娨娪娫娬娭娮娯娰娲娳娴娵娷娸娹娺娻娼娽娾娿婀婂婃婄婅婇婈婊婋婌婍婎婏婐婑婒婓婔婕婖婗婘婙婛婜婝婞婟婠婢婤婥婧婨婩婪婫婮婯婰婱婲婳婵婷婸婹婺婻婼婽婾媀媁媂媃媄媅媆媇媉媊媋媌媎媏媐媑媓媔媕媖媗媘媙媛媜媝媞媟媠媡媢媣媤媥媦媨媩媪媫媬媭媮媱媲媴媵媶媷媸媹媺媻媾嫀嫃嫄嫅嫆嫇嫈嫊嫍嫎嫏嫐嫑嫒嫓嫔嫕嫖嫘嫙嫚嫛嫜嫝嫞嫟嫠嫢嫣嫤嫥嫦嫧嫨嫪嫫嫬嫭嫮嫯嫱嫲嫳嫴嫶嫷嫸嫹嫼嫽嫾嬁嬂嬄嬅嬆嬊嬍嬎嬏嬐嬑嬒嬓嬔嬕嬖嬗嬘嬚嬛嬜嬞嬟嬠嬢嬥嬧嬨嬩嬫嬬嬮嬯嬱嬲嬳嬴嬵嬶嬷嬹嬺嬼嬽嬿孀孁孂孅孈孉孊孍孏孑孒孓孖孚孛孜孞孠孡孢孥孧孨孪孬孭孮孯孰孱孲孳孴孶孷孹孺宀宄宆宊宍宎宐宑宒宓宔宕宖実宥宧宨宩宬宭宯宱宲宷宸宺宻宼寁寅寈寉寊寋寍寎寏寐寔寕寖寗寘寙寚寛寜寣寤寪寭寮寯寰寱寲寴寷寽対尀専尃尌尐尓尕尗尛尜尞尡尢尣尥尦尧尨尩尪尫尬尭尮尯尰尲尳尴尵尶尹尻屃屄屇屌屐屒屔屖屗屘屙屚屝屟屣屦屧屪屫屮屰屲屳屴屵屶屷屸屺屻屼屽屾岀岃岄岆岇岈岉岊岋岌岍岎岏岐岑岒岓岕岘岙岚岜岝岞岟岠岢岣岤岥岦岧岨岪岫岬岮岯岰岱岲岴岵岶岷岹岺岻岼岽岾岿峀峁峂峃峄峅峆峇峈峉峊峋峌峍峎峏峐峑峒峓峔峕峖峗峘峙峚峛峜峞峟峠峢峣峤峥峧峨峩峪峫峬峮峱峲峳峵峷峸峹峺峼峾峿崀崁崂崃崄崅崆崈崉崊崋崌崏崐崒崓崕崘崚崛崜崝崞崟崠崡崣崤崥崦崧崨崪崫崮崯崰崱崲崳崴崵崶崷崸崹崺崻崼崽崾崿嵀嵁嵂嵃嵄嵅嵆嵇嵈嵉嵊嵋嵍嵎嵏嵑嵒嵓嵔嵕嵖嵘嵙嵚嵛嵜嵝嵞嵟嵠嵡嵢嵣嵤嵥嵦嵧嵨嵩嵪嵫嵬嵭嵮嵯嵰嵱嵲嵳嵴嵵嵶嵷嵸嵹嵺嵻嵿嶀嶂嶅嶆嶉嶊嶋嶌嶍嶎嶏嶐嶑嶒嶓嶕嶖嶙嶚嶛嶜嶝嶞嶟嶡嶣嶤嶥嶦嶩嶪嶫嶬嶭嶯嶰嶱嶲嶳嶴嶵嶶嶷嶻嶾嶿巀巁巂巃巄巅巆巇巈巉巌巎巏巐巑巓巕巙巚巛巜巟巣巤巪巬巭巯巳巶巸巺巻巼巽巿帄帇帉帊帍帎帏帑帒帓帔帗帙帛帞帟帠帡帢帣帤帧帨帩帪帯帰帱帲帴帵帷帹帺帻帼帾帿幁幂幄幆幈幉幊幋幍幎幏幐幒幖幛幜幞幠幡幤幥幦幧幨幪幭幮幯幰幱幵幷幺庀庁庂広庅庈庉庋庌庍庎庑庒庖庘庚庛庝庠庡庢庣庤庥庨庩庪庬庮庯庰庱庲庳庴庹庺庼庾庿廀廃廅廆廇廋廌廍廏廑廒廔廖廗廘廙廛廜廤廥廦廨廪廫廭廮廯廰廱廲廴廸廼廽廾廿弁弅弆弇弈弉弋弌弍弎弐弑弖弗弘弙弚弜弝弞弡弢弣弤弨弩弪弫弬弭弮弰弲弴弶弸弻弼弽弾弿彀彁彂彅彇彉彊彋彍彏彐彑彖彗彘彚彛彜彝彟彡彣彦彧彨彮彯彳彴彵彶彷彸彺彽彾徂徆徇徉徍徎徏従徔徕徖徚徛徜徝徟徢徣徤徥徦徨徫徬徭徯徰徱徲徳徴徶徸徺徻徼徾忀忁忂忄忇忈忉忊忋忎忏忐忑忒忓忔忕忖忚忛応忝忞忟忡忢忣忤忥忦忨忩忪忬忭忮忯忰忲忳忴忶忷忸忹忺忻忼忾怂怃怄怅怆怇怈怉怊怋怌怍怏怐怑怓怗怘怙怚怛怞怟怡怢怣怤怦怩怫怬怭怮怰怲怴怵怶怷怸怹怺怼怽怾怿恀恁恂恄恅恇恈恉恊恌恎恏恑恓恔恖恗恘恙恚恛恜恝恞恟恣恦恧恪恫恮恱恲恴恵恷恸恹恺恻恽恾恿悀悁悂悃悆悇悈悊悋悌悎悏悐悑悒悓悕悖悗悘悙悚悛悜悝悡悢悥悧悩悪悫悭悮悰悱悳悷悸悹悺悻悾悿惀惁惂惃惄惆惇惈惉惌惍惎惐惒惓惔惖惗惘惙惚惛惝惞惟惢惣惤惪惬惮惴惵惸惺惼惽惾惿愀愂愃愄愅愆愇愊愋愌愍愎愐愑愒愓愔愖愗愘愙愝愞愠愡愢愣愥愦愩愪愫愬愭愮愯愰愱愲愳愵愶愸愹愺愼慀慁慂慃慅慆慇慉慊慏慐慑慒慓慔慖慗慛慜慝慞慠慡慥慦慩慬慭慯慱慲慵慸慹慻慿憀憁憃憄憅憆憈憉憌憏憓憕憗憘憙憛憜憝憞憟憠憡憣憥憦憧憩憪憬憭憯憰憱憳憴憵憷憺憻憼憽憿懁懃懄懅懆懋懏懐懑懓懔懕懖懗懘懙懚懛懜懝懠懡懢懥懩懪懫懬懭懮懯懰懱懳懴懵懹懻懽懿戁戂戃戄戅戆戉戊戋戌戍戎戓戕戗戙戛戜戝戟戠戡戢戣戤戥戦戨戫戬戭戮戯戵戸戺戻戼戽戾扂扃扄扅扆扈扉扊扌扏扐払扖扗扙扚扜扝扟扠扡扢扤扥扦扨扪扱扲扴扵扷扸扺扻扽抁抂抃抅抆抇抈抉抋抌抍抎抏抐抔抙抜択抟抣抦抧抨抩抪抭抮抯抰抲抳抴抶抷抸抺抻抾抿拀拁拃拈拊拎拏拑拕拝拞拠拡拤拪拫拮拰拲拵拶拸拹拺拻拽挀挃挄挅挆挈挊挋挌挍挏挐挒挓挔挕挗挘挙挛挜挝挞挢挦挧挬挭挮挰挲挳挴挶挷挸挹挻挼挿捀捁捃捇捈捊捋捑捒捓捔捖捗捘捙捚捛捜捝捠捤捥捦捩捪捬捭捯捰捳捴捵捸捹捼捽捾捿掁掅掇掊掋掍掎掑掓掔掕掜掝掞掟掣掤掦掫掬掭掮掯掱掲掳掴掵掶掹掻掼掾掿揁揂揃揄揅揆揇揈揊揋揌揎揓揔揕揗揘揙揜揝揞揟揠揢揤揥揦揧揨揬揰揱揲揳揵揶揸揺揻揼揾揿搃搄搈搊搋搌搎搐搑搒搕搘搙搚搛搝搟搠搡搢搣搦搧搩搫搮搰搱搲搳搴搷搸搹搻搼搽搿摀摁摂摅摈摉摌摍摎摏摒摓摕摖摗摙摚摛摝摞摠摡摢摣摤摥摦摨摪摫摬摭摮摰摱摲摴摵摷摼摽摾摿撀撁撂撃撄撅撆撉撋撌撍撎撔撖撗撘撙撚撛撜撠撢撨撪撯撱撴撶撷撸撹撺撽擀擃擆擈擉擌擏擐擑擖擗擘擙擛擜擝擞擡擢擤擨擩擪擭擮擳擵擶擸擹擿攁攂攃攅攇攈攉攊攌攍攎攐攑攓攕攗攚攞攟攠攡攥攦攧攨攫攭攮攰攱攲攳攴攵攸攺攼攽敀敁敃敄敆敇敉敊敋敐敒敔敕敖敚敜敝敟敠敡敤敥敧敨敩敪敫敭敮敯敱敳敶敹敻敼敽敾敿斀斁斄斈斉斊斌斍斎斏斐斒斓斔斖斘斚斛斝斞斠斡斢斣斦斨斪斫斮斱斲斳斴斵斶斺斻斾斿旀旃旄旆旇旈旉旊旌旍旎旐旑旒旓旔旕旖旘旙旚旛旜旝旞旟旡旪旫旮旯旰旲旳旴旵旸旻旼旽旿昀昁昃昄昅昈昉昊昋昍昐昑昒昕昖昗昘昚昛昜昝昞昡昢昣昤昦昩昪昫昮昰昱昲昳昴昶昷昸昹昺昻昽昿晀晁晄晅晆晇晈晊晍晎晏晐晑晔晖晗晘晙晜晞晟晠晡晢晣晥晧晩晪晫晬晭晱晲晳晵晷晸晹晻晼晽晿暀暁暃暄暅暆暊暋暌暍暎暏暒暓暔暕暙暚暛暜暝暞暟暠暡暣暤暥暦暧暨暩暪暬暭暯暰暲暳暵暶暷暸暹暺暻暼暽暾暿曀曁曂曃曅曈曋曌曍曎曐曑曒曓曔曕曗曘曚曛曜曝曞曟曢曣曤曦曧曩曪曫曭曮曯曰曱曳曵曶曷曺曻曽朁朂朄朅朆朇朊朌朎朏朐朑朒朓朔朕朖朘朙朚朜朠朡朣朤朩札朰朲朳朷朸朹朻朼朾朿杁杄杅杇杊杋杌杍杒杓杔杕杗杘杙杚杛杝杞杢杣杤杦杧杩杪杫杬杮杲杳杵杶杷杸杹杺杻杼杽枀枂枃枅枆枇枈枊枋枌枍枎枏枑枒枓枔枖枘枙枛枞枟枠枡枤枥枦枧枨枩枬枭枮枰枲枳枵枸枹枺枻枼枽枾枿柀柁柂柃柅柆柇柈柉柊柋柌柍柎柕柖柗柘柙柚柛柝柞柟柡柢柣柤柦柧柨柩柪柫柭柮柯柰柲柶柷柸柹柼柽柾栀栂栃栄栆栉栊栌栍栎栐栒栔栕栘栙栚栛栜栝栟栠栣栤栥栦栧栨栩栫栬栭栮栯栱栲栳栴栵栶栺栻栾栿桀桁桄桇桉桊桋桍桎桏桓桔桕桖桗桘桙桚桛桜桝桞桟桠桡桢桤桧桪桫桬桭桯桰桲桳桴桵桷桸桹桻桼桽桾梀梂梃梄梇梈梉梊梋梌梍梎梏梐梑梒梓梕梙梚梛梞梠梡梣梤梥梩梪梫梬梮梱梴梵梶梷梸梹梺梻梼梽梾梿棁棂棃棅棆棇棈棌棎棏棐棑棓棔棙棛棜棝棞棢棣棤棥棦棨棩棪棫棬棭棯棰棳棴棷棸棹棻棼棽棾棿椀椁椂椃椄椆椇椈椉椊椋椌椐椑椓椔椕椖椘椙椚椛検椝椞椟椠椡椢椣椤椥椦椧椨椩椪椫椬椮椯椱椳椴椵椸椹椺椻椼椽楀楁楂楃楄楅楆楈楉楋楌楍楎楏楐楑楒楕楖楗楘楙楛楜楝楟楠楡楢楤楦楧楩楪楫楬楮楯楰楱楲楳楴楶楸楹楺楻楽楾楿榀榁榃榅榇榈榉榊榋榌榍榎榏榐榑榒榓榖榗榙榚榝榞榟榠榡榢榣榤榥榧榩榫榬榭榯榰榱榳榵榶榷榸榹榺榻榼榽榾槀槁槂槃槄槅槆槇槈槉槊槌槎槏槑槒槔槖槗様槙槚槛槜槝槞槟槠槡槢槣槥槦槩槪槬槭槯槰槱槲槴槵槷槸槹槺槻槾槿樀樃樄樆樇樈樉樋樌樍樎樏樐樒樔樕樖樗樘樚樛樜樝樥樦樨権樬樭樮樯樰樲樴樵樶樷樻樼樽樾橀橁橂橃橅橆橇橉橊橌橍橎橏橐橑橒橓橔橕橖橗橚橛橝橞橠橣橤橥橦橧橨橩橪橬橭橮橰橲橳橴橵橶橷橸橹橺橻橼橽橾橿檂檃檄檅檆檇檈檊檋檌檍檎檏檑檒檓檕檖檗檘檙檚檛檝檞檠檡檤檥檦檧檨檪檫檰檱檲檴檶檷檹檺檼檽檾檿櫀櫁櫂櫄櫆櫇櫉櫊櫋櫌櫎櫏櫐櫑櫒櫔櫕櫖櫗櫘櫙櫜櫡櫢櫣櫤櫦櫩櫭櫮櫯櫰櫲櫴櫵櫶櫷櫹櫺櫼櫽櫾櫿欀欁欂欃欆欈欉欋欌欎欔欕欗欙欚欛欜欟欤欥欦欨欩欪欫欬欭欮欯欰欱欳欴欶欷欸欹欻欼欿歀歁歂歃歄歅歆歈歊歋歍歏歑歒歓歔歕歖歗歘歙歚歜歝歞歠歨歩歫歬歭歮歯歰歱歳歵歶歺歽歾殁殂殄殅殆殇殈殌殍殎殏殐殑殒殓殔殕殗殙殚殛殜殝殟殠殡殣殥殦殧殩殪殬殱殳殶殸殹殽殾毂毃毄毇毈毉毋毌毎毐毑毓毖毗毘毚毜毝毞毟毠毢毣毤毥毦毨毩毪毭毮毰毱毲毳毴毵毶毷毸毹毺毻毼毽毾氀氁氃氄氅氆氇氉氋氍氎氐氒氕氖気氘氙氚氜氝氞氟氠氡氤氥氦氩氪氰氱氲氵氶氹氺氻氼氽氿汀汃汄汅汆汈汊汋汌汍汏汐汑汒汓汔汕汖汘汜汝汢汣汥汦汧汨汩汫汬汭汮汯汱汲汳汴汵汶汷汸汻汼汾汿沀沁沂沄沅沆沇沊沋沌沍沎沏沑沓沔沕沗沘沚沜沝沞沠沢沣沤沨沩沬沭沯沰沱沲沴沵沶沷沺泀泂泃泅泆泇泈泋泍泎泏泐泑泒泓泔泖泗泘泙泚泜泟泠泤泦泧泩泫泬泭泮泯泱泲泴泶泷泸泹泺泾泿洀洂洃洄洅洆洇洈洉洊洌洍洎洏洐洑洓洔洕洖洘洙洚洜洝洟洠洡洢洣洤洦洧洨洫洬洭洮洯洰洱洳洴洵洷洸洹洺洿浀浂浃浄浈浉浌浍浏浐浒浔浕浖浗浘浚浛浜浝浞浟浠浡浢浣浤浥浧浨浫浬浭浯浰浱浲浳浵浶浺浻浼浽浾涀涁涃涄涅涆涊涋涍涏涐涑涒涓涔涘涙涚涜涞涟涠涢涥涪涫涬涭涰涱涳涴涶涷涸涹涺涻涽涾涿淁淂淃淄淅淇淈淉淊淍淎淏淐淓淔淕淖淗淙淜淝淞淟淠淢淣淦淧淬淭淯淰淲淴淸淼淽淾淿渀渁渂渃渄渆渇済渉渋渌渍渎渏渑渒渓渕渘渚渜渝渞渟渥渧渨渪渫渭渮渰渱渲渳渵渶渷渹渻渼渽渿湀湁湂湄湅湆湇湈湉湌湍湎湏湐湑湒湓湔湕湗湙湚湛湜湝湟湠湡湢湣湤湥湦湨湩湪湫湬湭湮湰湱湲湳湴湵湶湷湸湹湺湻湽満溁溂溄溆溇溊溋溌溍溎溏溑溒溓溔溕溗溘溙溚溛溞溟溠溣溤溥溦溧溨溩溬溭溰溱溲溴溵溷溸溹溻溽溾溿滀滁滂滃滆滇滈滉滊滍滏滐滒滕滖滗滘滜滝滟滠滢滣滦滧滪滫滭滮滰滱滳滵滶滹滺滼滽漀漃漄漅漇漈漉漋漌漎漐漑漒漕漖漗漘漙漛漜漝漞漟漡漤漥漦漧漨漪漭漮漯漰漳漴漶漷漹漺漻漼漽潀潂潃潅潆潇潈潉潊潋潌潍潎潏潐潒潓潖潗潝潞潟潠潡潢潥潧潨潩潪潫潬潱潲潳潴潵潶潸潹潺潻潼潽潾澂澃澉澊澋澌澍澏澑澒澓澔澕澖澘澙澚澛澝澞澟澢澣澥澧澨澪澭澯澰澲澴澵澶澷澸澹澺澻澼澽澿濂濅濈濉濊濋濌濍濎濏濐濑濓濔濖濗濙濝濞濠濡濢濣濥濦濨濩濪濬濭濮濯濲濳濴濵濷濸濹濻濽瀀瀁瀄瀈瀊瀌瀍瀎瀐瀒瀓瀔瀖瀗瀙瀚瀛瀜瀞瀡瀢瀣瀤瀥瀩瀪瀫瀬瀭瀮瀯瀱瀳瀴瀵瀶瀷瀸瀹瀺瀻瀼瀽瀿灀灁灂灅灆灇灈灉灊灎灏灐灓灔灖灗灚灛灜灞灟灢灥灦灨灪灬灮灱灲灳灴灷灹灺灻炀炁炂炃炄炅炆炇炈炋炌炍炏炐炑炓炔炖炗炘炙炚炛炜炝炞炟炠炡炢炣炤炥炦炧炨炩炪炯炰炱炲炳炴炵炶炷炻炽炾炿烀烃烄烅烆烇烉烊烋烌烍烎烐烑烒烓烔烕烗烚烜烝烞烠烡烢烣烥烨烩烪烬烮烯烰烱烲烳烵烶烷烸烺烻烼烽烾烿焀焁焂焃焄焅焆焇焈焉焋焌焍焎焏焐焑焒焓焔焖焗焘焛焜焝焞焟焠焢焣焤焥焧焨焩焪焫焬焭焮焯焱焲焳焴焵焷焸焹焺焻焼焽焾焿煀煁煂煃煄煅煆煈煊煋煍煏煐煓煔煕煖煘煚煛煜煝煟煠煡煣煦煨煪煫煭煯煰煲煳煴煵煶煷煸煹煺煻煼煽煾煿熀熁熃熆熇熋熍熎熐熑熕熖熘熛熜熝熞熠熢熣熤熥熦熧熨熩熪熫熭熮熯熳熴熵熶熷熸熹熺熻熼熽熿燂燅燆燇燊燋燌燍燏燐燑燓燔燗燘燚燛燝燞燠燡燢燣燤燧燨燩燪燫燮燯燰燱燲燳燵燷燸燹燺燽燿爀爁爂爅爈爉爊爋爌爎爏爑爒爓爔爕爘爙爚爜爝爞爟爠爡爢爣爤爦爨爩爫爮爯爰爳爴爻爼爿牁牂牃牄牅牉牊牏牑牒牔牖牗牚牜牝牞牟牣牤牥牦牨牪牫牬牭牮牯牰牱牳牶牷牸牻牼牾牿犂犃犄犅犆犇犈犉犊犋犌犍犎犏犐犑犒犓犔犕犗犘犙犚犜犝犟犠犡犣犤犥犦犨犩犪犫犭犮犰犱犲犳犴犵犷犸犺犻犼犽犾犿狁狃狄狅狆狇狉狊狋狌狍狎狏狑狒狓狔狕狖狘狙狚狛狜狝狟狢狣狤狦狧狨狩狪狫狯狲狳狴狵狶狷狺狻狾狿猀猁猃猄猅猆猇猈猉猊猋猍猏猐猑猒猓猔猕猗猘猚猝猞猟猠猡猢猣猤猥猦猧猭猯猰猱猲猳猵猷猸猹猺猼猽獀獂獆獇獈獉獋獌獍獏獐獑獒獓獔獕獖獗獙獚獛獜獝獞獟獠獡獢獣獤獥獦獬獭獯獳獴獶獹獽獾獿玂玃玆玈玊玌玍玎玏玐玑玒玓玔玕玗玘玙玚玜玝玞玟玠玡玢玣玤玥玦玧玨玪玬玭玮玱玳玴玵玶玸玹玺玼玽玾玿珀珁珂珃珄珅珆珇珈珉珋珌珎珏珐珑珒珓珔珕珖珗珘珙珚珛珜珝珞珟珡珢珣珤珥珦珧珨珩珪珫珬珮珯珰珱珲珳珴珵珶珷珸珹珺珻珽珿琀琁琂琄琇琈琊琋琌琍琎琏琑琒琓琔琕琗琘琙琚琛琜琝琞琟琠琡琣琤琥琦琧琨琩琪琫琬琭琮琯琰琲琵琶琷琸琹琻琽琾瑀瑁瑂瑃瑄瑅瑆瑈瑉瑊瑌瑍瑎瑏瑐瑑瑓瑔瑕瑖瑗瑘瑙瑚瑛瑜瑝瑠瑡瑢瑥瑦瑧瑨瑫瑬瑭瑮瑱瑳瑴瑵瑶瑷瑸瑹瑺瑼瑾瑿璀璁璂璄璅璆璇璈璋璌璍璎璏璐璑璒璓璔璖璘璙璚璛璜璞璟璠璤璥璨璩璪璬璭璮璱璲璳璴璶璷璹璺璻瓀瓁瓂瓃瓆瓇瓈瓉瓋瓌瓍瓎瓐瓑瓒瓓瓖瓗瓘瓙瓝瓞瓟瓠瓡瓥瓧瓨瓩瓪瓫瓬瓭瓯瓰瓱瓲瓳瓴瓵瓸瓹瓺瓻瓼瓽瓾瓿甀甁甂甃甄甅甆甇甈甉甊甋甍"}

charclass_set = {k: set(v) for k, v in charclass_dict.items()}


@run_once('str')
def get_char_class(ch):
    for k, v in charclass_set.items():
        if ch in v:
            return k

    if re.match(r'\s', ch):
        return '空白'

    if re.match(r'\d', ch):
        return '数字'

    if re.match(r'[a-zA-Z]', ch):
        return '字母'

    if re.match(r'[\x00-\x7F]', ch):
        return '标点'

    return '其他字符'


def check_unclassified_chars(content, *, print_mode=0):
    """ 检查未被分类到的字符 """
    ct = Counter(content)
    ct2 = Counter()
    for ch, cnt in ct.most_common():
        cls = get_char_class(ch)
        if cls == '其他字符':
            if print_mode:
                print(ch, cnt)
            ct2[ch] = cnt
    return ct2


def get_charclass_num(content):
    """ 检查字符类型分布数量 """
    ct = Counter(content)
    ct2 = Counter()
    for ch, cnt in ct.most_common():
        cls = get_char_class(ch)
        ct2[cls] += cnt
    return ct2


def get_charclass_rate(content):
    """ 检查字符类型分布数量 """
    ct = get_charclass_num(content)

    # 这一类权重要翻倍
    for k in ['常用汉字', '繁体汉字', '偏僻汉字', '其他字符']:
        if k in ct:
            ct[k] *= 2

    total = sum(ct.values())
    ct2 = Counter()
    for k, v in ct.most_common():
        ct2[k] = v / total
    return ct2


def check_language_from_counter(ct):
    """ 检查语言类型 """
    # todo 这里的规则应该还可以再优化，以及不同的业务场景，这些规则应该也要再调整

    if '繁体汉字' in ct:
        if safe_div(ct['繁体汉字'], ct['常用汉字']) > 0.1:
            return '繁体中文'

    if ct['常用汉字'] + ct['偏僻汉字']:
        return '中文'

    if '其他字符' in ct:
        if ct['其他字符'] / ct['字母'] > 0.1:
            return '其他'

    if '字母' in ct:
        return '英文'

    return '其他'


def check_language_from_content(content):
    ct = get_charclass_num(content)
    return check_language_from_counter(ct)



# copyright (c) 2019 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import paddle
from paddle import nn


class CTCLoss(nn.Layer):
    def __init__(self, use_focal_loss=False, **kwargs):
        super(CTCLoss, self).__init__()
        self.loss_func = nn.CTCLoss(blank=0, reduction='none')
        self.use_focal_loss = use_focal_loss

    def forward(self, predicts, batch):
        if isinstance(predicts, (list, tuple)):
            predicts = predicts[-1]
        predicts = predicts.transpose((1, 0, 2))
        N, B, _ = predicts.shape
        preds_lengths = paddle.to_tensor([N] * B, dtype='int64')
        labels = batch[1].astype("int32")
        label_lengths = batch[2].astype('int64')
        loss = self.loss_func(predicts, labels, preds_lengths, label_lengths)
        if self.use_focal_loss:
            weight = paddle.exp(-loss)
            weight = paddle.subtract(paddle.to_tensor([1.0]), weight)
            weight = paddle.square(weight)
            loss = paddle.multiply(loss, weight)
        loss = loss.mean()
        return {'loss': loss}



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from paddle.optimizer.lr import LRScheduler


class CyclicalCosineDecay(LRScheduler):
    def __init__(self,
                 learning_rate,
                 T_max,
                 cycle=1,
                 last_epoch=-1,
                 eta_min=0.0,
                 verbose=False):
        """
        Cyclical cosine learning rate decay
        A learning rate which can be referred in https://arxiv.org/pdf/2012.12645.pdf
        Args:
            learning rate(float): learning rate
            T_max(int): maximum epoch num
            cycle(int): period of the cosine decay
            last_epoch (int, optional):  The index of last epoch. Can be set to restart training. Default: -1, means initial learning rate.
            eta_min(float): minimum learning rate during training
            verbose(bool): whether to print learning rate for each epoch
        """
        super(CyclicalCosineDecay, self).__init__(learning_rate, last_epoch,
                                                  verbose)
        self.cycle = cycle
        self.eta_min = eta_min

    def get_lr(self):
        if self.last_epoch == 0:
            return self.base_lr
        reletive_epoch = self.last_epoch % self.cycle
        lr = self.eta_min + 0.5 * (self.base_lr - self.eta_min) * \
                (1 + math.cos(math.pi * reletive_epoch / self.cycle))
        return lr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽，梁奕本（js去注释部分）
# @Email  : 877362867@qq.com, https://lyeebn.gitee.io/technology-shop/HeyBoss.html
# @Date   : 2023/10/20

from collections import Counter
import re
import textwrap

try:
    import jsbeautifier
except ModuleNotFoundError:
    pass


def __1_删注释功能():
    """
    用编译语法解析方式分析，清理JS中的注释，支持嵌套

    Usage:
    1、A simple function
    from pyxllib.text.jscode import dropJScomment
    dropJScomment(jsSourceCodeAsString)

    2、Object
    from pyxllib.text.jscode import JSParser
    js = JSParser(jsSourceCode)
    jsc = js.clearComment()
    """


def 删注释周围留空(c, arr):  # 应急
    arr.reverse()
    for i in arr:
        # 先用这个快速处理，有空再优化： https://blog.csdn.net/cooco369/article/details/82994932
        # c = c[:i].rstrip() + '%' + c[i:].lstrip()  # debug，定位
        有回车 = False
        l = r = 上一回车处 = i
        r = i + 1
        len_c = len(c)
        while len_c > l > 0:
            l -= 1
            ci = c[l]
            if ci == '\n':  # \r已统一为\n
                有回车 = True
            elif ci not in '\t \v':  # TODO 中文空格行不行
                break
        while r < len_c:
            ci = c[r]
            if ci in '\n\r':  # \r已统一为\n
                有回车 = True
                上一回车处 = r  # 保持缩进，但有 Bug 灵异，难道是 ?
            elif ci not in '\t \v':  # TODO 中文空格行不行
                break
            r += 1
            # print(r)
        c = c[:l + 1] + ('\n' if 有回车 else '') + c[上一回车处:]  # 有必要多留个空格吗
        # 已知 BUG：连续注释后的缩进无法保持，但不影响 JS 代码逻辑就是了
    return c


def 回溯区分正则除号(c):
    # 此函数参数需要提前处理：高偶合，非内聚
    # 原理：除号是二元运算符，其前面必有一个量：数值（可以是变量名或字符串字面量），这不太好穷举
    # 而正则为一字面量，前面可能必为某种运算符 = + ，或特殊符号：& | 逻辑? 括号 ( , 参数，[ { 对象 : ，或 ; 语句结束符，回车。有个坑：折行要注意。
    # 摆脱对正则的依赖，变成无第三无依赖
    # True 为 正则， False 为除号
    i = len(c)
    while i > 0:
        i -= 1
        ci = c[i]
        if ci in '\t \v':
            continue  # 暂时无法给出结论
        elif ci in '\n\r':  # 折行，暂时无法给出结论，其实可以与如上合并，区别可能是回车前可能是已省略的分号
            # return 回溯区分正则除号(c, i=i)  # 经验证，js 中 转义回车仅限于字符串内，运算符后是可折行的， \ 反而语法错误
            continue  # 经考虑，还是不必递归了
        elif ci in '{[(=,;?:&|!*~%^':  # todo ++ --  ，注：~ 为非预期操作，但因类型转换语法允许 https://developer.mozilla.org/zh-CN/docs/Web/JavaScript/Guide/Expressions_and_operators
            return True  # 正则
        elif ci == '+':  # todo ++ --
            i -= 1
            if i > 0 and c[i] == '+': return False  # 除号 # ++： / 前 为变量 为除尘
            return True  # 正则
        elif ci == '-':  # 要么  变量--，要反报错
            i -= 1
            if i > 0 and c[i] == '-': return False  # 除号 # ++： / 前 为变量 为除尘
            return True  # raise BaseException('/ 前单 -')  # 实测因类型转换，两边转数值类型，最多是 NaN，而不至于报错
        else:
            return False  # 除号，可能是变量，什么的
    return True  # 正则


class JSParser():
    def 普通引号(self, 号='"'):  # 号：开始结束定界符，下同
        while self.indexPointer < self.jsCodeLength:  # 正常会提前 return 见下
            self.indexPointer += 1
            si = self.jsSourceCode[self.indexPointer]
            self.jsWithoutComment += si  # 这应该是以下所有情况包括 else 都要的
            if si == 号:
                return
            elif si == '\\':  # 转义，提前吃掉， \n'"` 等
                # 即  r += '\\'
                self.indexPointer += 1
                if self.jsSourceCode[
                    self.indexPointer + 1] == 号:  # 超标报错正好，对应其语法错误，TODO 以后完善，实测 JS，  如果在引号中回车前没\反是其语法错误；就当源文件语法吧
                    pass
                self.jsWithoutComment += self.jsSourceCode[self.indexPointer]  # 转义后的字符好像都是要吃掉的，不用 if，待深入思考
            elif si == '\n':
                raise BaseException('原稿语法有误》缺右字符串定界：\a' + 号)

    def 反引号(self, 号='`'):
        while self.indexPointer < self.jsCodeLength:  # 正常会提前 return 见下
            self.indexPointer += 1
            si = self.jsSourceCode[self.indexPointer]
            self.jsWithoutComment += si  # 这应该是以下所有情况包括 else 都要的
            if si == 号:
                return
            elif si == '\\':
                # 即  r += '\\'
                self.indexPointer += 1
                if self.jsSourceCode[
                    self.indexPointer + 1] == 号:  # 超标报错正好，对应其语法错误，TODO 以后完善，实测 JS，  如果在引号中回车前没\反是其语法错误；就当源文件语法吧
                    pass
                self.jsWithoutComment += self.jsSourceCode[self.indexPointer]  # 转义后的字符好像都是要吃掉的，不用 if，待深入思考
            elif si == '$' and self.jsSourceCode[
                self.indexPointer + 1] == '{':  # 重要区别 。经实验 ${  后必有 } 否则报错：g = `2${2+3 7`
                self.indexPointer += 1
                self.jsWithoutComment += '{'
                self.反引号嵌套表达式允许多行注释()
            pass

    def 反引号嵌套表达式允许多行注释(self):  # 适合钻牛角尖，  TODO 用 堆栈 结构练手？  这好像可以 递归 main 了，差个 } return 吧
        while self.indexPointer < self.jsCodeLength:  # 正常会提前 return 见下
            self.indexPointer += 1
            si = self.jsSourceCode[self.indexPointer]
            if si == '}':
                self.jsWithoutComment += si
                return
            elif si in '"\'':  # 吃撑了的嵌套骚操作  si == '"' or si == "'"
                self.jsWithoutComment += si
                self.普通引号(号=si)
            elif si == "`":  # `，还能嵌套 String.raw` 算了，不玩了
                self.jsWithoutComment += si
                self.反引号()
            elif si == '/':  # 以下这一块逻辑 同 main ？
                self.反斜线然后呢()
                # if s[idx+1] == '*':  # 注意这里，嵌套了注释！超标正好报错，有效的 JS 代码 这里不会超标
                #     idx += 1
                #     多行注释()
                # elif s[idx+1] == '/':
                #     idx += 1
                #     单行注释()
                # else:
                #     r += s[idx]
            else:
                self.jsWithoutComment += si

    def 反斜线然后呢(self):  # 共用于【main】与 【反引号嵌套表达式允许多行注释】中的表达式
        偷看 = self.jsSourceCode[self.indexPointer + 1]
        if 偷看 == '/':
            self.单行注释()
            self.jsWithoutComment += self.注释占位
        elif 偷看 == '*':
            self.jsWithoutComment += self.注释占位
            self.indexPointer += 1
            self.多行注释()
        elif 回溯区分正则除号(
                self.jsWithoutComment):  # bool(re.search('[\n\r(\[\{=+,;&?|]([ \t]*|\\[\n\r])*[ \t]*$', self.jsWithoutComment)):  # def 区分正则或除号(): ←
            self.jsWithoutComment += '/'
            self.正则()
        else:
            self.jsWithoutComment += '/'
            # 除号不必特殊处理吧()
        # TODO，重要【严重】有没有可能是除号：q = 1 / 2 + /2/ // 除号、正则 、注释并存
        # 如果是除号，那之前应该有数值  \d)，不好判断，考虑到 JS 有些隐性的类型转换，如：'6' / 3
        # 如果是正则，往前回溯应该必有 = 或 + ，也可能在函数作为参数 （甚至可能还有变态 通过 \ 加回车，在正则前折行） re 如下

    def 源反引号(self, 号='`'):
        return self.反引号(号=号)  # TODO 先借用，也就差个 \ ，其它有什么区别待考虑
        pass

    def 单行注释(self):
        # while (s[idx] != '\n' or s[idx] != '\r') and idx < 长 :
        while (self.jsSourceCode[self.indexPointer] not in '\n\r') and self.indexPointer < self.jsCodeLength:
            self.indexPointer += 1
        self.注释location.append(len(self.jsWithoutComment))
        self.jsWithoutComment += '\n'  # 补回车于删注释点后，不能反了

    def 多行注释(self):
        self.indexPointer += 1  # 要吗？加过没
        while not (self.jsSourceCode[self.indexPointer] == '*' and self.jsSourceCode[
            self.indexPointer + 1] == '/'): self.indexPointer += 1
        self.indexPointer += 1
        # r += '\n'
        self.注释location.append(len(self.jsWithoutComment))

    def 正则(self):  # 正常的 JS 代码中 正则中无回车符
        while True:  # not (jsSourceCode[self.indexPointer + 1] == '/')
            self.indexPointer += 1  # TODO 放哪里
            si = self.jsSourceCode[self.indexPointer]
            self.jsWithoutComment += si
            if si == '/':
                return
            elif si in '\n\r':
                raise BaseException('正则还能折行？你是哪个老师教的')
            elif self.jsSourceCode[self.indexPointer] == '\\':
                self.indexPointer += 1  # TODO 放哪里
                self.jsWithoutComment += self.jsSourceCode[self.indexPointer]

    def __init__(self, jsSourceCode):
        # 就传源码吧，文件打开让用户自己去做，不然还得判断是文件名还是字符串，还得判断文件是否存在，还得依赖 os 包
        self.jsSourceCode = jsSourceCode.replace('\r', '\n') + '\n            '  # 防超标，为啥 .strip() 会异常
        self.jsCodeLength = len(jsSourceCode)
        self.indexPointer = -1  # 游标指针
        self.注释占位 = ''  # 生僻占位？以便后期删其前 \s
        self.注释location = []  # 记录当时 len(self.jsWithoutComment)，改天写吧
        self.jsWithoutComment = False  # init 先执行，所以不能执行调此类的其它函数，全写到一个这个函数里也麻烦，缩进了两级，如果要用方法可能重复执行，故用缓存法，

    def clearComment(self):  # 难道这个函数要在放类外吗（以便 init 能调用）
        if self.jsWithoutComment: return self.jsWithoutComment  # 已经求值过就用缓存，避免如下代码重复执行。这里一个变量两用：初始 bool 类型 False ，之后存清注释的str结果，类型改变，如果要让 GPT 改为 C++ 可能要注意一下，除了这里，别的地文都没有动态类型
        # 仅第一次计算，
        self.jsWithoutComment = ''
        while self.indexPointer < self.jsCodeLength:  # 要注意在哪里加 1 ，统一在处理开头处加吧，让当前处理的指标与相应字符初始一致，特殊情况再特殊加
            self.indexPointer += 1
            si = self.jsSourceCode[self.indexPointer]
            if si == '/':  # 正则、单/多行注释、除号，idea：注释前可能有多余的 \s，删注释时，可以留个 unicode 记号，到时用正则删
                self.反斜线然后呢()
                continue  # 其后的情况都要 r += si， TODO 合并
            elif si in '"\'':  # '  "        →1 si == '"' or si == "'"
                self.jsWithoutComment += si
                self.普通引号(号=si)
            elif si == "`":  # `
                self.jsWithoutComment += si
                self.反引号()
            elif si == "S" and self.jsCodeLength - self.indexPointer > 11 and self.jsSourceCode[
                                                                              self.indexPointer:self.indexPointer + 11] == 'String.raw`':  # String.raw` 元字符串
                self.jsWithoutComment += si
                self.indexPointer += 10;
                self.jsWithoutComment += 'tring.raw`'  # 分两步，以便上面那个能和其它情景合并
                self.源反引号()
            elif si == "\n":  # 压缩连续回车（空行）寻找一回车位置，这可能出现 BUG 吗？慎重，危险，不能出现于字符串，注释等中
                tmp = self.indexPointer
                while tmp < self.jsCodeLength:
                    tmp += 1
                    if self.jsSourceCode[tmp] == '\n':
                        self.indexPointer = tmp
                    elif self.jsSourceCode[tmp] in '\t ':
                        break
                    else:
                        break
                self.jsWithoutComment += si
            else:
                self.jsWithoutComment += si
            # 格式化字符串好像没什么特殊的
            pass
        self.jsWithoutComment = 删注释周围留空(self.jsWithoutComment, self.注释location).strip()
        # print(self.注释location)
        return self.jsWithoutComment


def remove_js_comments(jsSourceCode):  # 对外接口，将本来用得两行代码封装为一行
    js = JSParser(jsSourceCode)
    return js.clearComment()


def __2_类js的as处理功能():
    pass


airscript_head = r"""
// 0 基础组件代码（可以放在功能代码之前，也能放在最后面）

// 根据提供的 pattern 在 range 中寻找 cell
// 如果没有提供 range，默认在 ActiveSheet.UsedRange 中寻找
function findCell(pattern, range = ActiveSheet.UsedRange) {
    const cell = range.Find(pattern, range, xlValues, xlWhole)
    return cell
}

function levenshteinDistance(a, b) {
    const matrix = [];

    let i;
    for (i = 0; i <= b.length; i++) {
        matrix[i] = [i];
    }

    let j;
    for (j = 0; j <= a.length; j++) {
        matrix[0][j] = j;
    }

    for (i = 1; i <= b.length; i++) {
        for (j = 1; j <= a.length; j++) {
            if (b.charAt(i - 1) === a.charAt(j - 1)) {
                matrix[i][j] = matrix[i - 1][j - 1];
            } else {
                matrix[i][j] = Math.min(matrix[i - 1][j - 1] + 1, Math.min(matrix[i][j - 1] + 1, matrix[i - 1][j] + 1));
            }
        }
    }

    return matrix[b.length][a.length];
}

// 根据提供的 pattern 在 range 中寻找 column
// 如果没有提供 range，默认在 ActiveSheet.UsedRange 中寻找
function findColumn(pattern, range = ActiveSheet.UsedRange) {
    let cell = findCell(pattern, range);  // 首先尝试精确匹配
    if (!cell) {  // 如果精确匹配失败，尝试模糊匹配
        let minDistance = Infinity;
        let minDistanceColumn;
        for (let i = 1; i <= range.Columns.Count; i++) {
            let columnName = range.Cells(1, i).Value;
            let distance = levenshteinDistance(pattern, columnName);
            if (distance < minDistance) {
                minDistance = distance;
                minDistanceColumn = i;
            }
        }
        return minDistanceColumn;
    }
    if (cell) { return cell.Column }
}

// 根据提供的 pattern 在 range 中寻找 row
// 如果没有提供 range，默认在 ActiveSheet.UsedRange 中寻找
function findRow(pattern, range = ActiveSheet.UsedRange) {
    const cell = findCell(pattern, range)
    if (cell) { return cell.Row }
}

// 判断一个 cells 集合是否为空
function isEmpty(cells) {
    for (let i = 1; i <= cells.Count; i++) {
        if (cells.Item(i).Text) {
            return false;
        }
    }
    return true;
}

// 获取实际使用的区域
function getUsedRange(maxRows = 500, maxColumns = 100, startFromA1 = true) {
    /* 允许通过"表格上下文"信息，调整这里数据行的上限500行，或者列上限100列
        注意，如果分析预设的表格数据在这个限定参数内可以不改
        只有表格未知，或者明确数据量超过设置时，需要重新调整这里的参数
        调整的时候千万不要故意凑的刚刚好，可以设置一定的冗余区间
        比如数据说有4101条，那么这里阈值设置为5000也是可以的，比较保险。
    */

    // 默认获得的区间，有可能是有冗余的空行，所以还要进一步优化
    let usedRange = ActiveSheet.UsedRange;

    let lastRow = Math.min(usedRange.Rows.Count, maxRows);
    let lastColumn = Math.min(usedRange.Columns.Count, maxColumns);

    let firstRow = 1;
    let firstColumn = 1;

    // 找到最后一个非空行
    for (; lastRow >= firstRow; lastRow--) {
        if (!isEmpty(usedRange.Rows(lastRow).Cells)) {
            break;
        }
    }

    // 找到最后一个非空列
    for (; lastColumn >= firstColumn; lastColumn--) {
        if (!isEmpty(usedRange.Columns(lastColumn).Cells)) {
            break;
        }
    }

    // 如果表格不是从"A1"开始，找到第一个非空行和非空列
    if (!startFromA1) {
        for (; firstRow <= lastRow; firstRow++) {
            if (!isEmpty(usedRange.Rows(firstRow).Cells)) {
                break;
            }
        }

        for (; firstColumn <= lastColumn; firstColumn++) {
            if (!isEmpty(usedRange.Columns(firstColumn).Cells)) {
                break;
            }
        }
    }

    // 创建一个新的 Range 对象，它只包含非空的行和列
    let newUsedRange = ActiveSheet.Range(
        usedRange.Cells(firstRow, firstColumn),
        usedRange.Cells(lastRow, lastColumn)
    );

    return newUsedRange;  // 返回新的实际数据区域
}

// 将 Excel 日期转换为 JavaScript 日期
function xlDateToJSDate(xlDate) {
    return new Date((xlDate - 25569) * 24 * 3600 * 1000);
}

// 判断日期是否在本周
function isCurrentWeek(date) {
    const today = new Date();
    today.setHours(0, 0, 0, 0);  // 把时间设为午夜以准确地比较日期
    const firstDayOfWeek = new Date(today.setDate(today.getDate() - today.getDay()));
    const lastDayOfWeek = new Date(today.setDate(today.getDate() - today.getDay() + 6));
    return date >= firstDayOfWeek && date <= lastDayOfWeek;
}

// 判断日期是否在当前月份
function isCurrentMonth(date) {
    const currentDate = new Date();
    currentDate.setHours(0, 0, 0, 0);  // 把时间设为午夜stdcode以准确地比较日期
    return date.getMonth() === currentDate.getMonth() && date.getFullYear() === currentDate.getFullYear();
}

// 判断日期是否在下周
function isNextWeek(date) {
  const today = new Date();
  today.setHours(0, 0, 0, 0);  // 把时间设为午夜以准确地比较日期
  const nextWeek = new Date(today.getFullYear(), today.getMonth(), today.getDate() + 7);
  return date > today && date <= nextWeek;
}

// 判断日期是否在下个月
function isNextMonth(date) {
    const today = new Date();
    today.setHours(0, 0, 0, 0);  // 把时间设为午夜以准确地比较日期
    const nextMonth = new Date(today.getFullYear(), today.getMonth() + 1, 1);
    const endDateOfNextMonth = new Date(today.getFullYear(), today.getMonth() + 2, 0);
    return date >= nextMonth && date <= endDateOfNextMonth;
}
""".strip()


class AirScriptCodeFixer:
    @classmethod
    def fix_colors(cls, code_text):
        # 1 一些错误的颜色设置方法
        if re.search(r'(?<!\.)\b(Color.\w+)\b', code_text):
            return 0, code_text

        # 2 不能像vba那样，直接对颜色设置一个数值
        match = re.search(r'\.Color\s*=\s*(\d+)', code_text)
        if match:
            color_number = int(match.group(1))
            red = color_number % 256
            green = (color_number // 256) % 256
            blue = (color_number // 256 // 256) % 256
            rgb_format = f'RGB({red}, {green}, {blue})'
            code_text = code_text[:match.start(1)] + rgb_format + code_text[match.end(1):]

        # 3 一些错误的颜色设置方法，进行修正
        configs = {
            '红色': 'RGB(255, 0, 0)',
            '黄色': 'RGB(255, 255, 0)',
            '绿色': 'RGB(0, 255, 0)',
            '蓝色': 'RGB(0, 0, 255)',
            '灰色': 'RGB(128, 128, 128)',
            'red': 'RGB(255, 0, 0)',
            'yellow': 'RGB(255, 255, 0)',
            'green': 'RGB(0, 255, 0)',
            'blue': 'RGB(0, 0, 255)',
            'black': 'RGB(0, 0, 0)',
            'gray': 'RGB(128, 128, 128)',
            'grey': 'RGB(128, 128, 128)',
            'purple': 'RGB(128, 0, 128)',
            'pink': 'RGB(255, 192, 203)',
            'orange': 'RGB(255, 128, 0)',
        }

        def replace_color_fmt(m):
            t1, t2 = m.groups()
            t2 = t2.strip('"\'').lower()
            if t2 in configs:
                return f'{t1}{configs[t2]}'
            elif m2 := re.search(r'[a-fA-F0-9]{6}', t2):
                res = f'{t1}RGB({int(m2.group(0)[:2], 16)}, ' \
                      f'{int(m2.group(0)[2:4], 16)}, ' \
                      f'{int(m2.group(0)[4:], 16)})'
                return res
            return t1 + m.group(2)

        text = re.sub(r'''(\bColor\s*=\s*)(['"].+?['"])''', replace_color_fmt, code_text)

        # 4 经过优化仍无法修正的颜色问题
        if re.search(r'''\bColor\s*=\s*['"]''', text):
            # global count_target
            # ms = re.findall(r'''\bColor\s*=\s*(['"].+)''', text)
            # for m in ms:
            #     count_target[m] += 1
            return 0, text

        return 1, text

    @classmethod
    def fix_miscellaneous(cls, code_text):
        """ 修复其他各种杂项问题 """
        text = code_text

        # Cannot convert a Symbol value to a string, 一般是对Excel对象使用'+='运算报错
        text = re.sub(r'(\s+)((?:.+)Value2?)\s+(?:\+=)\s+(.+)', r'\1\2 = \2 + \3', text)  # 531条

        # 各种错误的接口调用形式
        text = text.replace('.Range.Find(', '.Find(')  # 8条

        # sort接口问题
        text = re.sub(r'(\.Sort\(.*?,\s+)(-1|0|false)\)', r'\g<1>2)', text)  # 328条

        # 做数据有效性的时候，有时候会有重复的引号嵌套
        text = re.sub(r'''(Formula\d:\s*')"(.+?)"''', r'\1\2', text)

        # 230907周四19:56，枚举值不用放在字符串中
        text = re.sub(r'''(['"`])(xlCellTypeVisible)\1''', r'\2', text)

        # 231106周一18:42，range的使用规范性
        text = re.sub(r'Range\(("|\')([A-Z]+|\d+)("|\')\)', r'Range(\1\2:\2\1)', text)

        return 1, text

    @classmethod
    def delete_error_record(cls, code_text):
        return 1, code_text

    @classmethod
    def check_assistant_content(cls, code_text):
        text = code_text

        global count_target
        pieces = re.findall(r'[a-zA-Z_\d\.]+\.Columns', text)
        count_target += Counter([x.strip() for x in pieces])

        # Columns前一般用ActiveSheet就行了

        return 1, text

    @classmethod
    def simplify_advtools(cls, code_text):
        """ 移除高级工具函数代码，用其他更简洁的方式取代 """
        text = code_text
        text = text.replace('getUsedRange()', 'ActiveSheet.UsedRange')
        text = re.sub(r'''findCell\(((['"]).+?\2)(, [a-zA-Z]+)?\)''',
                      r'ActiveSheet.UsedRange.Find(\1)', text)
        text = re.sub(r'''findColumn\(((['"]).+?\2)(, [a-zA-Z]+)?\)''',
                      r'ActiveSheet.UsedRange.Find(\1).Column', text)
        text = re.sub(r'''findRow\(((['"]).+?\2(, [a-zA-Z]+)?)\)''',
                      r'ActiveSheet.UsedRange.Find(\1).Row', text)

        return 1, text

    @classmethod
    def simplify_code(cls, code_text, indent=4):
        """ 代码简化，去掉一些冗余写法

        包括代码美化，默认缩进是4，但在训练阶段，建议默认缩进是2，
        """
        # 1 代码精简
        code_text = re.sub(r'Application\.(WorksheetFunction|ActiveWorkbook|ActiveSheet|Sheets|Range|Workbook)', r'\1', code_text)
        code_text = re.sub(r'Workbook\.(Sheets)', r'\1', code_text)
        code_text = re.sub(r'ActiveSheet\.(Range|Rows|Columns|Cells)', r'\1', code_text)
        code_text = re.sub(r'(\w+)\.(Row|Column)\s*\+\s*\1\.\2s\.Count\s*-\s*1', r'\1.\2End', code_text)
        code_text = re.sub(r'\bvar\b', 'let', code_text)
        code_text = code_text.replace('Sheets.Item(', 'Sheets(')
        code_text = re.sub(r'Application.Enum.\w+.(\w+)', r'\1', code_text)

        # 2 代码美化
        opts = jsbeautifier.default_options()
        opts.indent_size = indent
        code_text = jsbeautifier.beautify(code_text, opts)

        return 1, code_text.strip()

    @classmethod
    def simplify_code2(cls, code_text, indent=4):
        """ 有些规则可能在标注数据中想留着，但训练的时候想删除，则可以调用这个进一步级别的简化 """
        _, code_text = cls.simplify_code(code_text, indent)
        return code_text

    @classmethod
    def fix_stdcode(cls, code_text):
        """ 更智能的，缺什么组件才补什么组件 """
        # 1 检查依赖补充
        text = code_text
        _, text = cls.simplify_advtools(text)

        defined_vars = set(re.findall(r'(?:<=^|\b)(?:var|let|const|function)\s+(\w+)(?:\s+|\()', text))
        used_vars = set(re.findall(r'(?<!\.)\b(\w+)\b', text))

        # 2 提取js中的函数
        def extract_functions(code_string):
            pattern = r"(function\s+(\w+).+?^\})"
            matches = re.findall(pattern, code_string, re.MULTILINE | re.DOTALL)
            return {name: func for func, name in matches}

        js_funcs = extract_functions(airscript_head)

        # 3 补充缺失的定义
        pre_additional_code = []
        for name, code in {'xlDateToJSDate': '',
                           'isCurrentWeek': '',
                           'isCurrentMonth': '',
                           'isNextWeek': '',
                           'isNextMonth': '',
                           'usedRange': 'const usedRange = ActiveSheet.UsedRange;',
                           'headerRows': 'const headerRows = usedRange.Rows("1:1");',
                           'firstDataRow': 'const firstDataRow = headerRows.RowEnd + 1;',
                           'lastDataRow': 'const lastRow = usedRange.RowEnd;',
                           }.items():
            if name in used_vars and name not in defined_vars:
                if name in js_funcs:
                    code = js_funcs[name]
                if code:
                    pre_additional_code.append(code)
                    used_vars.remove(name)
                else:  # 有未定义就使用的变量，这条数据不要了
                    return 0, text
        else:
            # 还得再检查一波是不是有叫'xxxColumn'的变量未定义被使用
            logo = True
            for name in used_vars:
                if name.endswith('Column') and name not in defined_vars:
                    logo = False
                    break
            if logo and pre_additional_code:
                text = '\n'.join(pre_additional_code) + '\n' + text
            return 1, text

    @classmethod
    def pre_proc(cls, code_text):
        code_text = re.sub(r'^\\n', '', code_text, flags=re.MULTILINE)
        return 1, code_text

    @classmethod
    def fix_loc_head(cls, code_text):
        """ 修复定位头 """
        m1 = re.search(r'//\s*1([\.\s]+)定位', code_text)
        m2 = re.search(r'//\s*2([\.\s]+)业务功能', code_text)
        if not m1 and m2:
            code_text = '// 1' + m2.group(1) + '定位\n' + code_text
        return 1, code_text

    @classmethod
    def remove_stdcode(cls, code_text):
        """ 删除开头固定的组件头代码 """
        code_text = re.sub(r'(.*?)(//\s*1[\.\s]+定位)', r'\2', code_text, flags=re.DOTALL)
        code_text = re.sub(r'// 0 基础组件代码（可以放在功能代码之前，也能放在最后面）.+?$', '', code_text, flags=re.DOTALL)
        return 1, code_text

    @classmethod
    def fix_texts(cls, code_text):
        """ 修复文本中出现的关键词，描述 """
        s = code_text
        s = s.replace('<表格结构信息描述>', '表格摘要')
        s = s.replace('<孩子:表格摘要>', '表格摘要')
        return 1, s

    @classmethod
    def fix_base(cls, code_text):
        text = code_text
        for func in [
            cls.simplify_code,
            cls.fix_colors,
            cls.fix_miscellaneous,
            cls.advanced_remove_comments_regex,
        ]:
            status, text = func(text)
            if not status:
                return status, text
        return status, text

    @classmethod
    def fix_base2(cls, code_text):
        text = code_text
        for func in [
            cls.simplify_code,
            cls.fix_colors,
            cls.fix_miscellaneous,
        ]:
            status, text = func(text)
            if not status:
                return status, text
        return status, text

    @classmethod
    def fix_all(cls, code_text):
        old_text = code_text
        text = code_text
        for func in [
            cls.simplify_code,
            cls.fix_colors,
            cls.fix_miscellaneous,
            cls.fix_stdcode,
            # cls.advanced_remove_comments_regex,
        ]:
            status, text = func(text)
            if not status:
                return status, text
        # if text != old_text:
        #     bcompare(old_text, text)
        #     dprint()
        return status, text

    @classmethod
    def format_hanging_indent(cls, text):
        r""" 优化悬挂缩进的文本排版

        :param str text: 输入文本
        :return str: 优化后的文本

        >>> AirScriptCodeFixer.format_hanging_indent('const usedRange = getUsedRange();\\n        const headerRows = usedRange.Rows(\'1:1\');')
        'const usedRange = getUsedRange();\\nconst headerRows = usedRange.Rows(\'1:1\');'
        """
        lines = text.strip().split('\n')  # 去掉前后空行并分割成行
        first_line = lines.pop(0)  # 取出第1行
        remaining_text = '\n'.join(lines)  # 剩余行合并为一个字符串
        dedented_text = textwrap.dedent(remaining_text)  # 对剩余行进行反缩进处理

        return 1, first_line + '\n' + dedented_text  # 将处理后的剩余行和第1行拼接回去

    @classmethod
    def remove_comments_regex(cls, js_code):
        """ 这个代码功能并不严谨，只是一个临时快速方案 """
        js_code = re.sub(r'^\s*/\*.*?\*/\n?', '', js_code, flags=re.DOTALL | re.MULTILINE)
        js_code = re.sub(r'^\s*//.*\n?', '', js_code, flags=re.MULTILINE)

        # Removing multi-line comments
        js_code = re.sub(r'\s*/\*.*?\*/', '', js_code, flags=re.DOTALL)
        # Removing single-line comments
        js_code = re.sub(r'\s*//.*', '', js_code)
        return js_code

    @classmethod
    def advanced_remove_comments_regex(cls, js_code):
        # Regex to match strings, either single or double quoted
        string_pattern = r'(?:"[^"\\]*(?:\\.[^"\\]*)*"|\'[^\'\\]*(?:\\.[^\'\\]*)*\')'

        # Combined regex pattern to match strings or single/multi-line comments
        pattern = r'|'.join([
            string_pattern,  # match strings first to avoid removing content inside them
            r'\/\/[^\n]*',  # single line comments
            r'\/\*.*?\*\/'  # multi-line comments
        ])

        def replacer(match):
            # If the matched text is a string, return it unchanged
            if match.group(0).startswith(('"', "'")):
                return match.group(0)
            # Otherwise, it's a comment, so return an empty string
            return ''

        # Use re.sub with the replacer function
        return 1, re.sub(pattern, replacer, js_code, flags=re.DOTALL)

    @classmethod
    def remove_js_comment(cls, js_code):
        try:
            js_code2 = remove_js_comments(js_code)
        except BaseException as e:
            js_code2 = cls.remove_comments_regex(js_code)
        return 1, js_code2


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/11/05

""" 基于jieba库的一些文本处理功能 """

from collections import Counter
import re

from tqdm import tqdm
import pandas as pd

import jieba
import jieba.posseg as pseg

from pyxllib.prog.pupil import DictTool, run_once
from pyxllib.file.specialist import XlPath
from pyxllib.algo.stat import update_dataframes_to_excel


def jieba_add_words(words):
    for w in words:
        jieba.add_word(w)


def jieba_del_words(words):
    for w in words:
        jieba.del_word(w)


@run_once('str')
def jieba_cut(text):
    return tuple(jieba.cut(text))


@run_once('str')
def pseg_cut(text):
    return tuple(pseg.cut(text))


def _count_word_frequency(texts, function_word=True):
    """ 统计关键词出现频数 (主要是协助计算tf-idf)

    :param texts: 输入字符串列表
    :param function_word: 是否要统计虚词
    :return: 一个dict
        key: 分词名称
        values: [x, y]，x是出现总频数，y是这个词在多少篇文章中出现过

    >>> _count_word_frequency(['正正正正', '正反正', '反反反反'])
    {'正正': [1, 1], '反反': [2, 1]}

    原没有过滤词性的结果：{'正正': [2, 1], '正': [1, 1], '反正': [1, 1], '反反': [2, 1]}
    """

    d = dict()
    for text in tqdm(texts, '词频统计'):
        wordflags = list(pseg.cut(text))
        words = set()
        for word, flag in wordflags:
            # 虚词不做记录
            if (not function_word) and flag in ('uj', 'd', 'p', 'c', 'u', 'xc'):
                continue
            words.add(word)
            if word not in d:
                d[word] = [0, 0]
            d[word][0] += 1
        for word in words:
            d[word][1] += 1
    return d


def analyse_tf_idf(texts, outfile=None, sheet_name='tf-idf', *, function_word=True):
    """ 分析tf-idf值

    :param list[str] texts: 多份文件的文本内容
    :return: 一个DataFrame数据

    这个算法jieba可能有些自带库可以搞，但是自己写一下也不难啦
    注意我这里返回的tf-idf中，是放大了总频数倍的，这样显示的数值大一点，看起来舒服~
    """
    from math import log10

    frequency = _count_word_frequency(texts, function_word)
    DictTool.isub(frequency, [' ', '\t', '\n'])

    n = len(texts)
    sum_frequency = sum([v[0] for v in frequency.values()])

    li = []
    for k, v in frequency.items():
        idf = log10(n / v[1])
        # idf = 1
        li.append([k, v[0], v[0] / sum_frequency, v[1], idf, v[0] * idf])
    df = pd.DataFrame.from_records(li, columns=('词汇', '频数', '频率', '出现该词文章数', 'idf', 'tf-idf'))
    df.sort_values(by='tf-idf', ascending=False, inplace=True)

    if outfile:
        update_dataframes_to_excel(outfile, {sheet_name: df})

    return df


class TextClassifier:
    def __init__(self, texts=None):
        """ 文本分类器

        :param list[str] texts: 文本内容
        """

        self.texts = []
        self.tfidf = {}
        self.vecs = []  # 每份文本对应的向量化表达
        self.default_tfidf = 1  # 如果没有计算tf-idf，可以全部默认用权重1

        if texts:
            for text in texts:
                self.texts.append(text)

    def get_text_tf(self, text, *,
                    function_word_weight=0.2,
                    normalize=True,
                    ingore_words=(' ', '\t', '\n'),
                    add_flag=False):
        """ 这里可以定制提取text关键词的算法

        :param function_word_weight: 这里可以自定义功能性词汇权重，一般是设一个小数降低权重

        一般是定制一些过滤规则，比如过滤掉一些词性，或者过滤掉一些词
        """
        ct = Counter()

        # 1 初步的分词，以及是否要过滤虚词
        wordflags = list(pseg_cut(text))
        for word, flag in wordflags:
            if flag in ('uj', 'd', 'p', 'c', 'u', 'xc', 'x'):
                if add_flag:
                    ct[word + ',' + flag] += function_word_weight
                else:
                    ct[word] += function_word_weight
            else:
                if add_flag:
                    ct[word + ',' + flag] += 1
                else:
                    ct[word] += 1

        # 2 归一化一些词
        if normalize:
            ct2 = Counter()
            for k, v in ct.items():
                # 如果需要对一些词汇做归一化，也可以这里设置
                k = re.sub(r'\d', '0', k)  # 把数字都换成0
                ct2[k] += v
            ct = ct2

        # 3 过滤掉一些词
        if ingore_words:
            for k in ingore_words:
                if k in ct:
                    del ct[k]

        return ct

    def compute_tfidf(self, outfile=None, sheet_name='tf-idf', normalize=False, function_word_weight=0.2, add_flag=False):
        """ 重算tfidf表 """
        from math import log10

        # 1 统计频数和出现该词的文章数
        d = dict()
        for text in tqdm(self.texts, '词频统计'):
            ct = self.get_text_tf(text, normalize=normalize, function_word_weight=function_word_weight, add_flag=add_flag)
            for k, v in ct.items():
                if k not in d:
                    d[k] = [0, 0]
                d[k] = [d[k][0] + v, d[k][1] + 1]

        # 2 计算tfidf
        n = len(self.texts)
        sum_tf = sum([v[0] for v in d.values()])
        ls = []
        for k, v in d.items():
            idf = log10(n / v[1])
            # idf = 1
            ls.append([k, v[0], v[0] / sum_tf, v[1], idf, v[0] * idf])

        df = pd.DataFrame.from_records(ls, columns=('词汇', '频数', '频率', '出现该词文章数', 'idf', 'tf-idf'))
        df.sort_values(by='tf-idf', ascending=False, inplace=True)

        # 3 保存到文件
        if outfile:
            update_dataframes_to_excel(outfile, {sheet_name: df})

        self.tfidf = {row['词汇']: row['tf-idf'] for idx, row in df.iterrows()}
        self.default_tfidf = df.loc[len(df) - 1]['tf-idf']  # 最后条的权重作为其他未见词的默认权重

        return df

    def normalization(self, d):
        """ 向量归一化

        输入一个类字典结构表示的向量，对向量做归一化处理
        """
        length = sum([v * v for v in d.values()]) ** 0.5  # 向量长度
        return {k: v / length for k, v in d.items()}

    def get_text_vec(self, text):
        """ 获取文本的向量化表达

        :param str text: 文本内容
        """
        ct = self.get_text_tf(text)
        vec = {k: v * self.tfidf.get(k, self.default_tfidf) for k, v in ct.items()}
        vec = self.normalization(vec)
        return vec

    def compute_vecs(self):
        """ 重置向量化表达 """
        vecs = []
        for text in tqdm(self.texts, desc='query向量化'):
            vecs.append(self.get_text_vec(text))
        self.vecs = vecs
        return vecs

    def cosine_similar(self, x, y):
        """ 两个向量的余弦相似度，值越大越相似

        这里是简化的，只算两个向量的点积，请确保输入的都是单位长度的向量
        注意这里x和y都是稀疏矩阵的存储形式，传入的是dict结构
        """
        keys = x.keys() & y.keys()  # 求出x和y共有的键值
        return sum([x[k] * y[k] for k in keys])

    def find_similar_vec(self, x, maxn=10):
        """ 找与x最相近的向量，返回下标和相似度

        :pamra x: 待查找的对象
        :param maxn: 返回最相近的前maxn个对象
        """
        if isinstance(x, str):
            x = self.get_text_vec(x)

        # todo 使用并行计算？或者其实也可以向量化，但向量化是稀疏矩阵，挺占空间的
        sims = [(i, self.cosine_similar(x, v)) for i, v in enumerate(self.vecs)]
        sims.sort(key=lambda x: x[1], reverse=True)
        return sims[:maxn]

    def refine_vecs(self):
        """ 优化向量数据 """
        # 1 计算每个向量的长度
        vecs = []
        for vec in tqdm(self.vecs, '优化向量'):
            vec = [(k, v) for k, v in vec.items()]
            vec.sort(key=lambda x: x[1], reverse=True)
            vec2 = {}
            for k, v in vec:
                if v < 0.0001:
                    break
                vec2[k] = round(v, 4)
            vecs.append(vec2)

        self.vecs = vecs
        return self.vecs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/10/20

import re


class VBACodeFixer:
    @classmethod
    def simplify_code(cls, code_text):
        """ 代码简化，去掉一些冗余写法 """
        code_text = re.sub(r'ActiveSheet\.(Range|Rows|Columns|Cells)', r'\1', code_text)
        code_text = re.sub(r'(\w+)\.(Row|Column)\s*\+\s*\1\.\2s\.Count\s*-\s*1', r'\1.\2End', code_text)

        return 1, code_text.strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/02/21 11:07

"""
对PaddleOcr进行了更高程度的工程化封装
"""
import collections
import os
import sys
import re

import pandas as pd
import yaml
import shutil
import copy
import inspect
import math
import json

import numpy as np
from tqdm import tqdm

from pyxlpr.ppocr.tools.program import preprocess
from pyxlpr.ppocr.data import build_dataloader

from pyxllib.algo.geo import rect_bounds, ltrb2xywh
from pyxllib.file.specialist import XlPath, ensure_localfile, ensure_localdir
from pyxllib.cv.xlcvlib import xlcv
from pyxllib.prog.newbie import round_int


class PaddleOcrBaseConfig:
    """ paddle(ocr)标准配置文件的封装，为了简化配置方便自己使用，
    做了一个中间层组件，方便做一些统一的参数设置、修改
    """

    def __init__(self):
        self.cfg = {}

    def __1_config(self):
        """ 配置文件相关的功能 """
        pass

    def autoset(self):
        """ 这个接口方便写一些通用的配置 """

        x = self.cfg['Global']
        x['use_visualdl'] = True
        x['print_batch_step'] = 1000  # 这个单位是iter。原本很小2，我改成了100。但epoch很小的时候，每轮epoch也会输出。
        x['pretrained_model'] = None
        # 每隔多少次epoch，保存模型，原本默认是1200，这里故意设置的特别大，相当于不保存模型，需要的话手动设置补充。
        # 虽然没有固定间隔保存模型，但默认还是会根据eval，保存最优模型的
        x['save_epoch_step'] = 100000

        self.set_save_dir('models/' + inspect.stack()[3].function)

    def resume(self, train=False):
        """ 如果没有设置checkpoints，尝试加载best_accuracy或latest模型

        跟是否是Train模式有关，默认加载的模型会不一样
            train要加载latest，其他默认优先加载accuracy
        """
        if train:  # 用于模型训练时，应该是优先回复上一次的模型
            candidates = ['latest', 'best_accuracy']
        else:  # 用于其他场合，则应该是默认找最佳模型来使用
            candidates = ['best_accuracy', 'latest']

        for name in candidates:
            f = XlPath(self.cfg['Global']['save_model_dir']) / name
            if f.with_suffix('.pdparams').exists():
                self.cfg['Global']['checkpoints'] = f
                return

    def config_from_content(self, content):
        self.cfg = yaml.safe_load(content)
        self.autoset()
        return self.cfg

    def config_from_template(self, subpath):
        """
        :param subpath: 例如 'det/det_mv3_db'
        """
        f = os.path.join(sys.modules['pyxlpr.ppocr'].__path__[0], 'configs', subpath + '.yml')
        return self.config_from_content(XlPath(f).read_text())

    def set_save_dir(self, save_dir):
        """ 有很多个运行中文件的输出路径，可以统一到一个地方，并且只设置一次就够 """
        # self.d['Global']
        save_dir = XlPath(save_dir)
        x = self.cfg['Global']
        x['save_model_dir'] = save_dir  # train时模型存储目录
        x['save_inference_dir'] = save_dir / 'infer'  # export_model时存储目录
        # 这个选项暂时还不清楚具体作用，不知道是不是db专有的
        x['save_res_path'] = save_dir / 'predicts.txt'

    def set_simpledataset(self, mode, data_dir, label_file_list, ratio_list=None):
        """ paddle官方标准的SimpleDataset数据格式

        :param str mode: Train or Eval，设置训练集或者验证集
        :param PathLike data_dir: 数据所在根目录
        :param list label_file_list: 标注文件清单 [txtfile1, textfile2, ...]
            每个txtfile文件里的内容，每行是一张图的标注
            每行第1列是图片相对data_dir的路径，\t隔开，第2列是json.dumps的json标注数据
            json里有transcription字段存储文本内容，points存储四边形框位置
        :param list ratio_list: 只有一个label_file_list的时候，可以只输入一个数字，但最好统一按列表输入
            填写一个0~1.0的小数值，表示所取样本比例数
            这个paddle官方实现是随机取的，没有顺序规律
        """
        # 注意如果在SimpleDataSet、XlSimpleDataSet之间切换的话，有些字段格式是有区别的
        # 保险起见，就把self.cfg[mode]['dataset']整个重置了
        node = self.cfg[mode]['dataset']
        x = {'name': 'SimpleDataSet',
             'data_dir': XlPath(data_dir),
             'label_file_list': label_file_list}
        if ratio_list:
            x['ratio_list'] = ratio_list
        x['transforms'] = node['transforms']
        self.cfg[mode]['dataset'] = x

    def set_xlsimpledataset(self, mode, data_dir, data_list):
        """ 设置自己的XlSampleDataSet数据格式

        用于对各种源生的格式，在程序运行中将格式直接转为paddle的内存支持格式接口，从而不用重复生成冗余的中间数据文件
        目前最主要的是扩展了对xllabelme标注格式的支持，如labelme_det

        :param str mode: Train or Eval，设置训练集或者验证集
        :param PathLike data_dir: 数据所在根目录
        :param list data_list: 数据具体清单，每个条目都是一个字典
            [必填]type: 具体的数据格式，目前支持 labelme_det, icdar2015, refineAgree
                具体支持的方法，可以见XlSimpleDataSet类下前缀为from_的成员方法
            其他为选填字段，具体见from_定义支持的扩展功能，一般有以下常见参数功能
            [ratio] 一个小数比例，可以负数代表从后往前取
                一般用于懒得物理区分Train、Eval数据集的时候，在代码里用算法自动拆分训练、验证集

        """
        node = self.cfg[mode]['dataset']
        x = {'name': 'XlSimpleDataSet',
             'data_dir': XlPath(data_dir),
             'data_list': data_list}
        x['transforms'] = node['transforms']
        self.cfg[mode]['dataset'] = x

    @classmethod
    def _rset_posix_path(cls, d):
        from pathlib import Path

        if isinstance(d, list):
            for i, x in enumerate(d):
                if isinstance(x, (Path, XlPath)):
                    d[i] = x.as_posix()
                else:
                    cls._rset_posix_path(x)
        elif isinstance(d, dict):
            for k, v in d.items():
                if isinstance(v, (Path, XlPath)):
                    d[k] = v.as_posix()
                else:
                    cls._rset_posix_path(v)

    def rset_posix_path(self):
        """ 配置字典中，可能存在XlPath、Path类，需要递归统一修改为str类型来存储

        rset是递归设置的意思
        """
        d = copy.deepcopy(self.cfg)
        self._rset_posix_path(d)
        return d

    def write_cfg_tempfile(self):
        """ 存储一个文件到临时目录，并返回文件路径 """
        p = XlPath.tempfile('.yml')
        # TODO 写入文件前，会把配置里 XlPath全部转为 as_poisx 的str
        self._rset_posix_path(self.cfg)
        p.write_yaml(self.cfg)
        return str(p)

    def add_config_to_cmd_argv(self):
        """ 把配置参数加入命令行的 -c 命令中 """
        sys.argv = sys.argv + ['-c', self.write_cfg_tempfile()]

    def set_iter_num(self, num):
        """ 按迭代数设置训练长度

        paddle的配置源生并不支持按iter来统计训练长度，
        要通过batch_size_per_card和数据量，来反推epoch_num需要设置多少

        注意要先设置好数据集，再继续迭代数哦！
        """
        config, device, logger, _ = preprocess(from_dict=self.rset_posix_path(), use_visualdl=False)
        train_dataloader = build_dataloader(config, 'Train', device, logger)
        per_epoch_iter_num = len(train_dataloader)  # 每个epoch的迭代数
        self.cfg['Global']['epoch_num'] = math.ceil(num / per_epoch_iter_num)

    def __2_main(self):
        """ 一些脚本功能工具 """
        pass

    def train(self, resume=False):
        from pyxlpr.ppocr.tools.train import main

        if resume:
            self.resume(True)
        config, device, logger, vdl_writer = preprocess(is_train=True, from_dict=self.rset_posix_path())
        main(config, device, logger, vdl_writer)

    def eval(self, resume=True, *, dataset_mode='Eval'):
        """
        :param dataset_mode: 使用的数据集，默认是Eval，也可以用Train
        """
        from pyxlpr.ppocr.tools.eval import main

        if resume:
            self.resume()

        config, device, logger, vdl_writer = preprocess(from_dict=self.rset_posix_path())
        for k in ['name', 'data_dir', 'data_list']:
            config['Eval']['dataset'][k] = config[dataset_mode]['dataset'][k]
        metric = main(config, device, logger)
        return metric

    def infer_det(self, resume=True):
        from pyxlpr.ppocr.tools.infer_det import main

        if resume:
            self.resume()
        config, device, logger, vdl_writer = preprocess(from_dict=self.rset_posix_path())
        main(config, logger)

    def export_model(self, resume=True):
        from pyxlpr.ppocr.tools.export_model import main

        if resume:
            self.resume()
        config, device, logger, vdl_writer = preprocess(from_dict=self.rset_posix_path())
        main(config, logger)

    def __3_pretrained(self):
        """ 使用预训练模型相关配置的封装 """

    @classmethod
    def get_pretrained_model_backbone(cls, name):
        """ 只拿骨干网络的权重 """
        local_file = XlPath.userdir() / f'.paddleocr/pretrained/{name}.pdparams'
        url = f'https://paddle-imagenet-models-name.bj.bcebos.com/dygraph/{name}.pdparams'
        ensure_localfile(local_file, url)
        return local_file.parent / local_file.stem  # 省略.pdparams后缀

    @classmethod
    def get_pretrained_model_ppocr(cls, name):
        local_dir = XlPath.userdir() / f'.paddleocr/pretrained/{name}'
        url = f'https://paddleocr.bj.bcebos.com/dygraph_v2.0/ch/{name}.tar'
        ensure_localdir(local_dir, url, wrap=-1)
        return local_dir / 'best_accuracy'  # ppocr训练好的ocr模型

    def set_pretrained_model_backbone(self, name):
        path = self.get_pretrained_model_backbone(name)
        self.cfg['Global']['pretrained_model'] = path

    def set_pretrained_model_ppocr(self, name):
        path = self.get_pretrained_model_ppocr(name)
        self.cfg['Global']['pretrained_model'] = path

    def set_pretrained_infer_model(self, local_dir, url):
        """ 自己扩展的一个配置参数，metric的时候用 """
        local_dir = XlPath.userdir() / f'.paddleocr/pretrained_infer/{local_dir}'
        path = ensure_localdir(local_dir, url, wrap=-1)
        self.cfg['Global']['pretrained_infer_model'] = path

    def set_pretrained_model(self, pretrained, models):
        """ 对上述功能进一步封装，简化高层接口配置时的代码复杂度

        :param bool|int pretrained:
            0 不使用预训练权重
            1 使用骨干网络权重
            2 使用完整的ppocr权重
            3 之前定制训练过的最好的模型
        :param models: pretrained为1、2时加载的模型
        """
        if pretrained == 1:
            self.set_pretrained_model_backbone(models[0])
        elif pretrained == 2:
            self.set_pretrained_model_ppocr(models[1])
        elif pretrained == 3:
            self.cfg['Global']['pretrained_model'] = self.cfg['Global']['save_model_dir'] / 'best_accuracy'

    def __call__(self, *args, **kwargs):
        # 让fire库配合return self不会报错
        pass


class XlDetText(PaddleOcrBaseConfig):
    """ 检测模型专用配置
    """

    def autolabel(self, datadir, *, model_type=0, **kwargs):
        """ 预标注检测、识别

        TODO model_type在det1_mobile的时候，默认设为2？

        """
        pocr = self.build_ppocr(model_type, **kwargs)
        pocr.ocr2labelme(datadir, det=True, rec=True)

    def set_deploy_args_det(self):
        """ 检测模型在部署时候的参数，不一定跟eval一样
        换句话说，eval本来就应该尽量贴近真实部署的配置参数

        由于很多文本检测的配置文件，在eval时有些配置跟部署不同，这里按照deploy的情况自动进行调整

        当然，有些配置，如果eval效果确实比deploy来的好，可以考虑deploy采用eval的配置方式
        """
        for x in self.cfg['Eval']['dataset']['transforms']:
            if 'DetResizeForTest' in x:
                x['DetResizeForTest'] = {'limit_side_len': 960, 'limit_type': 'max'}

    def det1_mobile_init(self, *, pretrained=2):
        """
        官方实验:ic15, train1000+val500张, batch_size_per_card=8, epoch=1200
            也就是总训练量120w，除batchsize，是15万iter
            按照核酸的实验，每iter耗时大概是0.4秒，实验总用时15iter/3600*0.4约等于17小时

        batchsize=8，hesuan训练过程占用显存 6.7G
            以后有其他实验数据，会尝试都覆盖上，但记忆中差不多都是消耗这么多

        这个部署文件共 3M

        TODO datalist不只一个的相关功能，还没有进行测试，但问题应该不大
        """
        # 1 加载基础的配置
        cfg = self.config_from_template('det/ch_ppocr_v2.0/ch_det_mv3_db_v2.0')
        self.set_pretrained_model(pretrained, ['MobileNetV3_large_x0_5_pretrained', 'ch_ppocr_mobile_v2.0_det_train'])
        self.set_deploy_args_det()

        # 2 预训练权重也提供一个部署模型，供后续metric分析
        infer_model_url = 'https://paddleocr.bj.bcebos.com/dygraph_v2.0/ch/ch_ppocr_mobile_v2.0_det_infer.tar'
        self.set_pretrained_infer_model('ch_ppocr_mobile_v2.0_det_infer', infer_model_url)

    def det1_server_init(self, *, pretrained=2):
        """
        训练显存 10.2 G

        这个部署文件共 47M
        """
        # 1 加载基础的配置
        cfg = self.config_from_template('det/ch_ppocr_v2.0/ch_det_res18_db_v2.0')
        self.set_pretrained_model(pretrained, ['ResNet18_vd_pretrained', 'ch_ppocr_server_v2.0_det_train'])
        self.set_deploy_args_det()

        # 2 预训练权重也提供一个部署模型，供后续metric分析
        infer_model_url = 'https://paddleocr.bj.bcebos.com/dygraph_v2.0/ch/ch_ppocr_server_v2.0_det_infer.tar'
        self.set_pretrained_infer_model('ch_ppocr_server_v2.0_det_infer', infer_model_url)

    def det2_init(self, *, pretrained=1):
        """ 2021.9.7发布的PP-OCRv2
        但是我还没有试跑过，不确定我这样配置是对的

        220223周三18:11，跑通了，但还没有完全对，metric结果有点奇怪，摸不着头脑
        """
        cfg = self.config_from_template('det/ch_PP-OCRv2/ch_PP-OCRv2_det_distill')
        if pretrained:
            x = cfg['Architecture']['Models']

            # self.set_pretrained_model_ppocr('ch_PP-OCRv2_det_distill_train')
            x['Student']['pretrained'] = self.get_pretrained_model_backbone('MobileNetV3_large_x0_5_pretrained')
            # x['Student']['pretrained'] = self.get_pretrained_model_ppocr('ch_PP-OCRv2_det_distill_train')
            x['Teacher']['pretrained'] = self.get_pretrained_model_ppocr('ch_ppocr_server_v2.0_det_train')

        self.set_deploy_args_det()

        infer_model_url = 'https://paddleocr.bj.bcebos.com/PP-OCRv2/chinese/ch_PP-OCRv2_det_infer.tar'
        self.set_pretrained_infer_model('ch_PP-OCRv2_det_infer', infer_model_url)

        return self

    def build_ppocr(self, model_type=2, **kwargs):
        """ 获得部署用的接口类
        导出部署模型，并加载

        :param model_type:
            0，原始的PaddleOcr
            1，配置文件自带的部署文件（需要配置Global.pretrained_infer_model参数）
            2，finetune后的模型
        :param kwargs: 可以增加一些检测模型的配置参数，比如常用的 det_db_unclip_ratio=1.5
        """
        from pyxlpr.paddleocr import PaddleOCR

        if model_type == 0:
            ppocr = PaddleOCR.build_ppocr(**kwargs)
        elif model_type == 1:
            d = self.cfg['Global']['pretrained_infer_model']
            if not d:
                return {}
            ppocr = PaddleOCR.build_ppocr(det_model_dir=d, **kwargs)
        else:
            self.export_model(True)
            ppocr = PaddleOCR.build_ppocr(det_model_dir=self.cfg['Global']['save_inference_dir'], **kwargs)

        return ppocr

    def _build_dataset(self, config, logger, dataset_mode='Eval'):
        from pyxlpr.ppocr.data import build_dataset
        # 注意这里数据集切换方法跟PaddleOCRConfig.eval有点不太一样，因为部署操作要连transforms一起改掉
        src = config[dataset_mode]['dataset']
        config['Eval']['dataset'] = {'name': src['name'],
                                     'data_dir': src['data_dir'],
                                     'data_list': src['data_list'],
                                     'transforms': [{'DetLabelEncode': None}]}
        dataset = build_dataset(config, 'Eval', logger)
        return config, dataset

    def eval_deploy(self, model_type=2, dataset_mode='Eval', **kwargs):
        ppocr = self.build_ppocr(model_type, **kwargs)
        config, device, logger, vdl_writer = preprocess(from_dict=self.rset_posix_path())
        config, dataset = self._build_dataset(config, logger, dataset_mode)
        metric = ppocr.det_metric(dataset)
        logger.info(str(metric))
        return metric

    def metric(self, *, print_mode=False):
        """ 得到一个综合的测评结果，一般如下：
                   type train_dataset eval_dataset
           ①PaddleOCR*      32.35*56    100.0*190
           ②pretrained      17.57*43      50.0*22
          ③pretrained*     17.57*184     50.0*192
             ④finetune      93.05*49     100.0*20
            ⑤finetune*     93.05*173    100.0*164

        有几条规律
        1、精度②=③，速度③>②。如果精度不同，很可能官方给的预训练模型和部署文件有问题
        2、精度④=⑤，速度⑤>④。如果精度不同，可能eval和部署阶段的图片处理方式不同
            det就存在这个问题，处理后的图片尺寸不同，通过set_deploy_args_det修复了
        3、去掉上述两个eval阶段的，其实就是①③⑤三个模型间的比较
            ①PaddleOCR直接给的一条龙准备好的模型，一般要高于③开源模型训练效果，低于⑤定制化的效果
            即精读：③<①<⑤
        """
        import pandas as pd
        from pyxllib.algo.stat import xlpivot

        # 1 收集各模型结果
        eval_list = []

        def core(title, eval_func):
            for dataset in ['a、Train', 'b、Eval']:
                m = eval_func(dataset[2:])  # m, metric
                m = {k: (round_int(v) if k in ('fps', 'total_frame') else round(v * 100, 2)) for k, v in m.items()}
                m2 = {'model_type': title, 'dataset': dataset}
                m2.update(m)
                eval_list.append(m2)

        core('①PaddleOCR*', lambda m: self.eval_deploy(model_type=0, dataset_mode=m))
        core('②pretrained', lambda m: self.eval(resume=False, dataset_mode=m))
        core('③pretrained*', lambda m: self.eval_deploy(model_type=1, dataset_mode=m))
        core('④finetune', lambda m: self.eval(resume=True, dataset_mode=m))
        core('⑤finetune*', lambda m: self.eval_deploy(model_type=2, dataset_mode=m))

        # 2 最后的统计表
        df = pd.DataFrame.from_records(eval_list)
        outfile = self.cfg['Global']['save_model_dir'] / f'results/metric.html'
        os.makedirs(outfile.parent, exist_ok=True)

        def func(items):
            x = items.iloc[0]
            return f'{x["precision"]:.0f}，{x["recall"]:.0f}，{x["hmean"]:.2f}，{x["fps"]}'

        df2 = xlpivot(df, ['model_type'], ['dataset', 'total_frame'], {'precision，recall，hmean，fps': func})
        stat_html = df2.to_html()
        stat_html = stat_html.replace('<th></th>', f'<th>{sys.argv[2]}</th>', 1)
        outfile.write_text(stat_html)

        if 'metric' in sys.argv:
            print(df2)
            return

        if print_mode:
            print(df2)

        return df

    def create_visual_results(self, *, model_type=2, max_samples=None, **kwargs):
        """ 将可视化结果生成到目录下

        :param max_samples: 限制生成的可视化图片上限，有时候只需要看少量样本

        【算法流程】基本思路是将数据转成coco格式后，使用coco的功能接口来实现，考验我以前接口好不好用的时候到了
        1、初始化指定的ppocr
        2、用ppocr生成一套检测结果
        3、和gt对比，生成一套coco数据
        4、生成coco可视化结果
        5、生成coco的数据分析表格
        """
        import PIL.Image
        from pyxlpr.data.coco import CocoGtData, CocoMatch

        ppocr = self.build_ppocr(model_type, **kwargs)
        for dataset_mode in ['Train', 'Eval']:  # 训练集和验证集结果都生成，放在两个不同目录
            gt = {'images': [],
                  'annotations': [],
                  'categories': CocoGtData.gen_categories(['text'])}
            dt = []
            k = 1

            config, device, logger, vdl_writer = preprocess(from_dict=self.rset_posix_path())
            config, dataset = self._build_dataset(config, logger, dataset_mode)
            out_dir = self.cfg['Global']['save_model_dir'] / f'results/{dataset_mode}'
            data_dir = self.cfg['Eval']['dataset']['data_dir']
            for img_id, x in enumerate(dataset, start=1):
                if max_samples and img_id > max_samples:
                    break

                # 1 拷贝图片数据到相对目录
                src_img_path = x['img_path']
                rel_img_path = XlPath(src_img_path).relpath(data_dir)
                dst_img_path = out_dir / rel_img_path
                os.makedirs(dst_img_path.parent, exist_ok=True)
                if not dst_img_path.is_file():
                    shutil.copy2(src_img_path, dst_img_path)

                # 2 生成对应图片的gt标注数据
                w, h = PIL.Image.open(str(dst_img_path)).size
                gt['images'].append(CocoGtData.gen_image(img_id, rel_img_path, h, w))
                for p in x['polys']:
                    gt['annotations'].append(
                        CocoGtData.gen_annotation(id=k, image_id=img_id, points=p, text=x['texts']))
                    k += 1

                # 3 生成dt标注数据
                img = xlcv.read_from_buffer(x['image'])
                for p in ppocr.ocr(img, rec=False):
                    dt.append({'image_id': img_id, 'category_id': 1, 'segmentation': np.array(p).reshape([1, -1]),
                               'bbox': ltrb2xywh(rect_bounds(p)), 'score': 1.0})

            cm = CocoMatch(gt, dt)
            cm.to_labelme_match(out_dir, segmentation=True)
            cm.to_excel(out_dir / 'cocomatch.xlsx')

    def __config_demo(self):
        """ 常用的配置示例 """

    def set_xllabelme_dataset(self, data_dir, ratio_list):
        """ 设置xllabelme格式的文字检测标注数据

        我自设的一种简单的数据集范式

        :param data_dir: 数据根目录
        :param list[float, float] ratio_list: 训练集、验证集所需的比例
            可以取负数，表示从后往前取；底层设置了随机数种子，每次取得具体文件是固定的。
            数据集较少的话，一般是推荐 [0.9, -0.1]，较多的话可以 [0.8, -0.2]
        """
        self.set_xlsimpledataset('Train', data_dir, [{'type': 'labelme_det', 'ratio': ratio_list[0]}])
        self.set_xlsimpledataset('Eval', data_dir, [{'type': 'labelme_det', 'ratio': ratio_list[1]}])

    def det1_mobile_raw(self):
        """ paddle源生格式的配置示例 """
        self.det1_mobile_init(pretrained=2)  # 基础配置
        self.set_save_dir('train/det1_mobile_raw')  # 模型保存位置
        self.set_simpledataset('Train', 'data', ['data/ppdet_train.txt'])
        self.set_simpledataset('Eval', 'data', ['data/ppdet_val.txt'])
        self.set_iter_num(150000)
        return self

    def det1_mobile(self):
        """ labelme标注格式的检测训练 """
        self.det1_mobile_init(pretrained=2)  # 基础配置
        self.set_save_dir('train/det1_mobile')  # 模型保存位置
        self.set_xllabelme_dataset('data', [0.9, -0.1])  # 设置数据集
        self.set_iter_num(150000)  # 设置迭代轮次
        return self

    def det1_server(self):
        self.det1_server_init(pretrained=2)  # 基础配置
        self.set_save_dir('train/det1_server')  # 模型保存位置
        self.set_xllabelme_dataset('data', [0.9, -0.1])  # 设置数据集
        self.set_iter_num(150000)  # 设置迭代轮次
        return self


class XlRec(PaddleOcrBaseConfig):
    """ 识别模型专用配置
    """

    def stat_texts(self, xllabelme_data_dir, *, ref_dict='ppocr_keys_v1.txt'):
        """ 检查标注的句子、字符出现情况 statistics texts

        :param xllabelme_data_dir: xllabelme格式的标注数据所在目录
        :param ref_dict: 参考字典文件
        """
        from collections import Counter
        from pyxllib.algo.pupil import ValuesStat
        from pyxllib.algo.stat import dataframes_to_excel
        from pyxlpr.ppocr.utils import get_dict_content

        root = XlPath(xllabelme_data_dir)
        outfile = root.parent / 'stat_texts.xlsx'

        # 1 读取数据
        sentances_counter = Counter()  # 每句话的内容，和相同话出现的次数
        for f in root.rglob('*.json'):
            for sp in f.read_json()['shapes']:
                attr = json.loads(sp['label'])
                if 'text' in attr:
                    text = attr['text']
                    sentances_counter[text] += 1

        # 2 统计 sentances 每句话出现频率, words 每个单词出现频率, chars 每个字符出现频率
        chars_counter = Counter()
        words_counter = Counter()
        for sentance, cnt in sentances_counter.items():
            for word in sentance.split():  # 目前先按空格分开，虽然严格来说，对于中文情况，要用结巴分词处理更准确
                words_counter[word] += cnt
            for ch in sentance:  # 统计每个字符出现次数，包括空格
                chars_counter[ch] += cnt

        # 3 转df
        char_dict = set(get_dict_content(ref_dict).splitlines())
        ls = []
        new_chars = []
        for char, cnt in chars_counter.most_common():
            ls.append([char, cnt, '' if char in char_dict else 'True'])
            if char not in char_dict and char != ' ':
                new_chars.append(char)
        chars_df = pd.DataFrame.from_records(ls, columns=['char', 'count', 'new_char'])

        words_df = pd.DataFrame.from_records(words_counter.most_common(), columns=['word', 'count'])
        sentances_df = pd.DataFrame.from_records([[sentance, cnt, len(sentance)]
                                                  for sentance, cnt in sentances_counter.most_common()],
                                                 columns=['sentance', 'count', 'length'])

        # 计算不同长度句子的分布规律
        ct = Counter()
        lengths = []
        for _, row in sentances_df.iterrows():
            ct[row['length']] += row['count']
            lengths += [row['length']] * row['count']  # 这个实现不是那么得优雅，但如果要兼容ValuesStat只能先这样处理
        # ct = sentances_df.groupby('length').sum().to_dict()['count']
        max_len = max(sentances_df['length'])
        sentances_length_df = pd.DataFrame.from_records([[i, ct.get(i, 0)] for i in range(max_len + 1)],
                                                        columns=['length', 'count'])

        # 4 频数规律计算
        def summary(title, vals):
            msg = ValuesStat(vals).summary(['g', '.2f', '.2f', 'g', 'g'])
            # print(msg)
            return [title] + re.findall(r':\s+(\S+)', msg)

        print('【stat_texts】')
        print(f'输出文件：{outfile.as_posix()}')

        print(f'不在字典中的{len(new_chars)}个字符：' + ''.join(new_chars))

        ls = [
            summary('字符频数', chars_df['count']),
            summary('词组频数', words_df['count']),
            summary('句子频数', sentances_df['count']),
            summary('句子长度', lengths),
        ]
        df = pd.DataFrame.from_records(ls, columns=['title', '总和', '均值标准差', '总数', '最小值', '最大值'])
        print(df)

        # 5 存储分析表
        sheets = {'字符': chars_df, '词组': words_df,
                  '句子': sentances_df, '句子长度': sentances_length_df}
        dataframes_to_excel(outfile, sheets)

    def create_recdata(self, src, dst, *, print_mode=True, recreate=False):
        """ 从xllabelme标注的格式，生成到paddle支持的识别数据格式；提取出供文本识别模型训练的文本行数据

        :param src: xllabelme_data_dir
        :param dst: 目标存储位置的根目录
        :param recreate: 如果目标目录存在，将其删除，重新生成

        注意：本套生成方法仅供参考，这套处理目前不是那么泛用
        """
        # 0
        src, dst = XlPath(src), XlPath(dst)
        if recreate and dst.is_dir():
            dst.delete()  # 如果已有，将其删除

        # 1 生成图片
        chars = set()
        labels1, labels2 = [], []
        for f in tqdm(list(src.rglob('*.json')), desc='提取文本行数据', disable=not print_mode):
            data = f.read_json()
            impath = f.parent / data['imagePath']
            im = xlcv.read(impath)
            for i, sp in enumerate(data['shapes'], start=1):
                # a组，提取文本行的时候，按外接矩形框截取
                name = f'imgs/{f.stem}_r{i:03}.jpg'
                text = json.loads(sp['label'])['text']
                chars |= set(text)
                xlcv.write(xlcv.get_sub(im, sp['points']), dst / name)
                labels1.append(f'{name}\t{text}')

                # b组，提取文本行的时候，进行仿射变换矫正
                name = f'imgs/{f.stem}_w{i:03}.jpg'
                xlcv.write(xlcv.get_sub(im, sp['points'], warp_quad=True), dst / name)
                labels2.append(f'{name}\t{text}')

        # 2 字典文件
        chars -= set(' \n\t')  # 要去掉空格等字符
        (dst / 'char_dict.txt').write_text('\n'.join(sorted(chars)))

        # 3 标注数据
        (dst / 'labels_rect.txt').write_text('\n'.join(labels1))
        (dst / 'labels_warp.txt').write_text('\n'.join(labels2))
        (dst / 'labels_total.txt').write_text('\n'.join(labels1 + labels2))

        return self

    def set_rec_dataset(self, data_dir, label_file_list):
        """ 设置识别数据集

        :param data_dir: 数据所在根目录
        :param list[str|list] label_file_list: 标注文件清单
            str，标注文件的相对路径
            list[str, float]，除了str描述标注文件路径，还有个ratio值配置选取样本的比例

        TODO 想做设置的集成，但目前还没想到好的设计方式，可以自己手动拆分数据，并在autoset中配置，也不会很麻烦
        """

        # self.cfg['Train']['dataset']['data_dir'] = Paths.eleclabel / 'recdata'
        # self.cfg['Train']['dataset']['label_file_list'] = [Paths.eleclabel / 'recdata/labels_ab.txt']
        # self.cfg['Eval']['dataset']['data_dir'] = Paths.eleclabel / 'recdata'
        # self.cfg['Eval']['dataset']['label_file_list'] = [Paths.eleclabel / 'recdata/labels_ab.txt']

        raise NotImplementedError


class XlCls:
    """ 分类模型，这个是基本使用源生的paddlepaddle，没有使用有个更强的paddleclas """


class XlOcr:
    """ 封装了文字技术体系，检测识别的一些标准化处理流程 """

    def __init__(self, root):
        self.root = XlPath(root)  # 项目根目录

    def step1_autolabel(self):
        """ 预标注检测、识别 """

    def step2_refinelabel(self):
        """ 人工手动优化label标注 """

    def step3_det(self):
        """ 训练检测模型 """
